#include <algorithm>
#include "CClustering.h"

// -----------------------------------------------------------------------------
// C C L U S T E R I N G
// -----------------------------------------------------------------------------

// twoMeans
void CClustering::twoMeans(std::vector<CKMeansElement>& aList) {
  CVector<float> aCentroid0;
  CVector<float> aCentroid1;
  int aNum0, aNum1;
  twoMeans(aList,aCentroid0,aCentroid1,aNum0,aNum1);
}

void CClustering::twoMeans(std::vector<CKMeansElement>& aList, CVector<float>& aCentroid0, CVector<float>& aCentroid1, int& aNum0, int& aNum1) {
  int aDim = aList[0].mData->data().size();
  // Initialize labels
  CVector<float> aCenterSum0(aDim);
  CVector<float> aCenterSum1(aDim);
  do {
    aCenterSum0 = 0.0; aCenterSum1 = 0.0;
    aNum0 = 0; aNum1 = 0;
    for (unsigned int i = 0; i < aList.size(); i++) {
      if (aList[i].mIgnoreInClustering) continue;
      aList[i].mLabel = (char)(1.0*(aNum0+1)/(aNum0+aNum1+1)+NMath::random());
      aList[i].mLower0 = 0.0;
      aList[i].mLower1 = 0.0;
      aList[i].mUpper = 10e20;
      if (aList[i].mLabel == 0) {
        aCenterSum0 += aList[i].mData->data();
        aNum0++;
      }
      else {
        aCenterSum1 += aList[i].mData->data();
        aNum1++;
      }
    }
  }
  while (aNum0 == 0 || aNum1 == 0);
  aCentroid0 = aCenterSum0;
  aCentroid0 *= 1.0/aNum0;
  aCentroid1 = aCenterSum1;
  aCentroid1 *= 1.0/aNum1;
  float aCenterDistance = 0.0;
  for (int m = 0; m < aDim; m++) {
    float temp = aCentroid0(m)-aCentroid1(m);
    aCenterDistance += temp*temp;
  }
  aCenterDistance = 0.5*sqrt(aCenterDistance);
  int aChanged;
  int aIterations = 0;
  // K-Means iterations
  do {
    aChanged = 0;
    for (unsigned int i = 0; i < aList.size(); i++) {
      CKMeansElement& aElement = aList[i];
      if (aElement.mIgnoreInClustering) continue;
      // Check if relabeling assigns the point to a closer cluster center
      if (aElement.mUpper <= aCenterDistance) continue;
      if (aElement.mLabel == 0) {
        // How far is the other center at least? Closer than the maximum distance to the current center?
        if (aElement.mUpper > aElement.mLower1) {
          // Improve approximation of the upper bound by computing dist0 exactly
          float dist0 = 0.0;
          CVector<float>& aVec = aElement.mData->data();
          for (int m = 0; m < aDim; m++) {
            float temp = aVec(m)-aCentroid0(m);
            dist0 += temp*temp;
          }
          dist0 = sqrt(dist0);
          aElement.mLower0 = dist0;
          aElement.mUpper = dist0;
          // Is it still possible that the point fits better to the new cluster?
          if (dist0 > aElement.mLower1) {
            // Improve also the approximation of the lower bound by computing dist1 exactly
            float dist1 = 0.0;
            for (int m = 0; m < aDim; m++) {
              float temp = aVec(m)-aCentroid1(m);
              dist1 += temp*temp;
            }
            dist1 = sqrt(dist1);
            aElement.mLower1 = dist1;
            // Now we can finally decide
            if (dist1 < dist0) {
              aCenterSum0 -= aVec; aNum0--;
              aCenterSum1 += aVec; aNum1++;
              aElement.mLabel = 1;
              aElement.mUpper = dist1;
              aChanged++;
            }
          }
        }
      }
      // The same in green....
      else {
        if (aElement.mUpper > aElement.mLower0) {
          float dist1 = 0.0;
          CVector<float>& aVec = aElement.mData->data();
          for (int m = 0; m < aDim; m++) {
            float temp = aVec(m)-aCentroid1(m);
            dist1 += temp*temp;
          }
          dist1 = sqrt(dist1);
          aElement.mLower1 = dist1;
          aElement.mUpper = dist1;
          if (dist1 > aElement.mLower0) {
            float dist0 = 0.0;
            for (int m = 0; m < aDim; m++) {
              float temp = aVec(m)-aCentroid0(m);
              dist0 += temp*temp;
            }
            dist0 = sqrt(dist0);
            aElement.mLower0 = dist0;
            if (dist0 < dist1) {
              aCenterSum1 -= aVec; aNum1--;
              aCenterSum0 += aVec; aNum0++;
              aElement.mLabel = 0;
              aElement.mUpper = dist0;
              aChanged++;
            }
          }
        }
      }
    }
    if (aChanged > 0) {
      // Update centroids
      CVector<float> aOldCentroid0(aCentroid0);
      CVector<float> aOldCentroid1(aCentroid1);
      aCentroid0 = aCenterSum0;
      aCentroid0 *= 1.0/aNum0;
      aCentroid1 = aCenterSum1;
      aCentroid1 *= 1.0/aNum1;
      // Update centroid distances
      float distc0 = 0.0;
      float distc1 = 0.0;
      aCenterDistance = 0.0;
      for (int m = 0; m < aDim; m++) {
        float temp = aCentroid0(m)-aOldCentroid0(m);
        distc0 += temp*temp;
        temp = aCentroid1(m)-aOldCentroid1(m);
        distc1 += temp*temp;
        temp = aCentroid0(m)-aCentroid1(m);
        aCenterDistance += temp*temp;
      }
      distc0 = sqrt(distc0);
      distc1 = sqrt(distc1);
      aCenterDistance = 0.5*sqrt(aCenterDistance);
      // Update upper and lower bounds
      for (unsigned int i = 0; i < aList.size(); i++) {
        CKMeansElement& aElement = aList[i];
        if (aElement.mIgnoreInClustering) continue;
        // Cluster centers moved. Worst case for lower bounds: they directly approach the point
        aElement.mLower0 -= distc0;
        aElement.mLower1 -= distc1;
        // Worst case for upper bound: the center moves straight away from the point
        if (aElement.mLabel == 0) aElement.mUpper += distc0;
        else aElement.mUpper += distc1;
      }
    }
    aIterations++;
  }
  while (aChanged > 0 && aIterations <= 5);
  if (aNum0 <= 1 || aNum1 <= 1) {
    std::cout << "Warning: Cluster of size 0 or 1" << std::endl;
  }
}

void CClustering::kMeans(std::vector<CSimpleKMeansElement>& aList, int K, CVector<float>* aCentroid, int* aNum) {
  int aDim = aList[0].mData->size();
  CVector<int> aBestLabels(aList.size());
  float bestEnergy = 1e20f;
  // Run with different initializations
  for (int t = 0; t < 10; t++) {
    // Initialize labels
    CVector<float>* aCenterSum = new CVector<float>[K];
    for (int k = 0; k < K; k++)
      aCenterSum[k].setSize(aDim);
    bool aNullFound;
    do {
      for (int k = 0; k < K; k++) {
        aCenterSum[k] = 0.0;
        aNum[k] = 0;
      }
      for (unsigned int i = 0; i < aList.size(); i++) {
        aList[i].mLabel = (int)(K*NMath::random());
        if (aList[i].mLabel >= K) aList[i].mLabel = K-1;
        aCenterSum[aList[i].mLabel] += (*aList[i].mData);
        aNum[aList[i].mLabel]++;
      }
      aNullFound = false;
      for (int k = 0; k < K; k++)
        if (aNum[k] == 0) { aNullFound = true; break; }
    }
    while (aNullFound);
    int aChanged;
    int aIterations = 0;
    // K-Means iterations
    do {
      for (int k = 0; k < K; k++) {
        aCentroid[k] = aCenterSum[k];
        aCentroid[k] *= 1.0/aNum[k];
      }
      aChanged = 0;
      for (unsigned int i = 0; i < aList.size(); i++) {
        CSimpleKMeansElement& aElement = aList[i];
        CVector<float>& aVec1 = (*aElement.mData);
        float bestDist = 10e20;
        int bestLabel;
        for (int k = 0; k < K; k++) {
          float dist = 0;
          CVector<float>& aVec2 = aCentroid[k];
          for (int i = 0; i < aDim; i++) {
            float temp = aVec1(i)-aVec2(i);
            dist += temp*temp;
          }
          if (dist < bestDist) {
            bestDist = dist;
            bestLabel = k;
          }
        }
        if (bestLabel != aElement.mLabel) {
          int oldLabel = aElement.mLabel;
          aCenterSum[oldLabel] -= aVec1;
          aNum[oldLabel]--;
          aCentroid[oldLabel] = aCenterSum[oldLabel];
          if (aNum[oldLabel] <= 0)
            aNum[oldLabel] = 1;
          aCentroid[oldLabel] *= 1.0/aNum[oldLabel];
          aCenterSum[bestLabel] += aVec1;
          aNum[bestLabel]++;
          aCentroid[bestLabel] = aCenterSum[bestLabel];
          aCentroid[bestLabel] *= 1.0/aNum[bestLabel];
          aElement.mLabel = bestLabel;
          aChanged++;
        }
      }
      aIterations++;
    }
    while (aChanged > 0 && aIterations <= 10);
    if (aChanged > 0)
      for (int k = 0; k < K; k++) {
        aCentroid[k] = aCenterSum[k];
        aCentroid[k] *= 1.0/aNum[k];
      }
    // Compute energy
    float aEnergy = 0.0f;
    for (unsigned int i = 0; i < aList.size(); i++) {
      CSimpleKMeansElement& aElement = aList[i];
      CVector<float>& aVec1 = (*aElement.mData);
      CVector<float>& aVec2 = aCentroid[aElement.mLabel];
      float dist = 0.0f;
      for (int i = 0; i < aDim; i++) {
        float temp = aVec1(i)-aVec2(i);
        dist += temp*temp;
      }
      aEnergy += dist;
    }
    if (aEnergy < bestEnergy) {
      for (unsigned int i = 0; i < aList.size(); i++)
        aBestLabels(i) = aList[i].mLabel;
      bestEnergy = aEnergy;
    }
  }
  // Return labels and centroids with best energy
  for (int k = 0; k < K; k++) {
    aCentroid[k] = 0.0;
    aNum[k] = 0;
  }
  for (unsigned int i = 0; i < aList.size(); i++) {
    CSimpleKMeansElement& aElement = aList[i];
    int k = aElement.mLabel = aBestLabels(i);
    aCentroid[k] += (*aElement.mData);
    aNum[k]++;
  }
  for (int k = 0; k < K; k++) 
    if (aNum[k] > 0) aCentroid[k] *= 1.0f/aNum[k];
}

#if 0
// bic
float CClustering::bic(std::vector<CSimpleKMeansElement>& aList, int K, CVector<float>* aCentroid, int* aNum) {
  if ((int)aList.size() <= K) return -1e20;
  int aDim = aList[0].mData->size();
  float var = 0.0f;
  for (unsigned int i = 0; i < aList.size(); i++) {
    CVector<float>& aVec1 = *aList[i].mData;
    CVector<float>& aVec2 = aCentroid[aList[i].mLabel];
    for (int i = 0; i < aDim; i++) {
      float temp = aVec1(i)-aVec2(i);
      var += temp*temp;
    }
  }
  int num = aList.size();
  var *= 1.0f/(aList.size()-K);
  float aBic = //-0.5f*num*log(2.0f*NMath::Pi)
               -0.5f*aDim*num*log(var)
               //-0.5f*num
               +0.5f*K
               //-num*log(num)
               -0.5f*aDim*K*log(num)
               //-0.5f*log(num)
               ;
  for (int k = 0; k < K; k++)
    aBic += aNum[k]*log(aNum[k]);
  return aBic;
}

#else

// bic
float CClustering::bic(std::vector<CSimpleKMeansElement>& aList, int K, CVector<float>* aCentroid, int* aNum, float aLambda, float aScale) {
  if ((int)aList.size() <= K) return -1e20;
  int aDim = aList[0].mData->size();
  float var = 0.0f;
  for (unsigned int i = 0; i < aList.size(); i++) {
    CVector<float>& aVec1 = *aList[i].mData;
    CVector<float>& aVec2 = aCentroid[aList[i].mLabel];
    for (int j = 0; j < aDim; j++) {
      float temp = aVec1(j)-aVec2(j);
      var += temp*temp;
    }
  }
  var *= 1.0f/(aList.size()-K);
  // Average distortion per point
  float aBic = -sqrt(var);
  // Penality for every cluster, particularly for small clusters
  for (int k = 0; k < K; k++) {
    float n = aNum[k]*aScale;
    aBic -= aLambda*log(n)/n;
  }
  return aBic;
}

#endif

// xMeans
void CClustering::xMeans(std::vector<CSimpleKMeansElement>& aList, int& K, CVector<float>*& aCentroid, int*& aNum, float aLambda, float aScale) {
  // Compute Bic of the trivial single cluster solution
  int aDim = aList[0].mData->size();
  for (unsigned int i = 0; i < aList.size(); i++)
    aList[i].mLabel = 0;
  CVector<float> aMean(aDim,0);
  for (unsigned int i = 0; i < aList.size(); i++)
    aMean += *aList[i].mData;
  aMean *= 1.0f/aList.size();
  int aNum1 = aList.size();
  float bestBic = bic(aList,1,&aMean,&aNum1,aLambda,aScale);
  std::cout << "1: " << bestBic;
  int bestK = 1;
  // Test other models
  for (int k = 2; k <= K; k++) {
    CVector<float>* aCentroidk = new CVector<float>[k];
    int* aNumk = new int[k];
    kMeans(aList,k,aCentroidk,aNumk);
    float newBic = bic(aList,k,aCentroidk,aNumk,aLambda,aScale);
    std::cout << " " << k << ": " << newBic;
    if (newBic > bestBic) {
      bestBic = newBic;
      bestK = k;
    }
    delete[] aCentroidk;
    delete[] aNumk;
  }
  std::cout << std::endl;
  // Run kMeans with best model for final result
  aCentroid = new CVector<float>[bestK];
  aNum = new int[bestK];
  kMeans(aList,bestK,aCentroid,aNum);
  K = bestK;
}


