#ifndef CClusteringH
#define CClusteringH

#include <vector>
#include <queue>
#include <CVector.h>
#include <NMath.h>

class CDataElement {
public:
  CDataElement() {};
  CDataElement(CVector<float>& aData) { mData = aData;};
  CVector<float>& data() {return mData;}
  float mScore;
  CVector<float> mData;
};

struct pointer_less_than {
const bool operator()(const CDataElement *a, const CDataElement *b) const {
  return a->mScore < b->mScore;
}
};

class CKMeansElement {
public:
  CKMeansElement() {};
  CKMeansElement(CDataElement* aData) {mData = aData; mIgnoreInClustering = false;};
  CKMeansElement(CDataElement* aData, bool aIgnoreInClustering) {mData = aData; mIgnoreInClustering = aIgnoreInClustering;};

  CDataElement* mData;
  char mLabel;
  bool mIgnoreInClustering;
  // Lower and upper bounds for fast k-means
  float mLower0, mLower1;
  float mUpper;
};

class CSimpleKMeansElement {
public:
  CSimpleKMeansElement() {};
  CSimpleKMeansElement(CVector<float>* aData) {mData = aData;}
  CVector<float>* mData;
  int mLabel;
  float mFloatDummy;
  int mIntDummy;
};

class CClustering {
public:
  CClustering() {}
  ~CClustering() {}
  void twoMeans(std::vector<CKMeansElement>& aList, CVector<float>& aCentroid0, CVector<float>& aCentroid1, int& aNum0, int& aNum1);
  void twoMeans(std::vector<CKMeansElement>& aList);
  void kMeansRandomInit(std::vector<CSimpleKMeansElement>& aList, int K);
  void kMeansCentroidInit(std::vector<CSimpleKMeansElement>& aList, int K, CVector<float>* aCentroid);
  void kMeans(std::vector<CSimpleKMeansElement>& aList, int K, CVector<float>* aCentroid, int* aNum);
  void xMeans(std::vector<CSimpleKMeansElement>& aList, int& K, CVector<float>*& aCentroid, int*& aNum, float aLambda, float aScale);

protected:
  // Model log likelihood for model selection
  float bic(std::vector<CSimpleKMeansElement>& aList, int K, CVector<float>* aCentroid, int* aNum, float aLambda, float aScale);
};

#endif

