#include "CGraph.h"
#include <NMath.h>

// lanczos ---------------------------------------------------------------------
void lanczos(CGraph& aGraph, CVector<float>& X) {
  // Problem size
  int n = aGraph.size();
  X.setSize(n);
  // Number of eigenvectors
  int nev = 1;
  // Standard eigenvalue problem
  int ido = 0;
  char bmat[2] = "I";
  // Searching the largest eigenvalues
  char which[3] = "LA";
  // Precision
  double tol = 0.0;
  double *resid;
  resid = new double[n];
  // Number of Ritz vectors
  int ncv = 5*nev;
  if (ncv>n) ncv = n;
  double *v;
  int ldv = n;
  v = new double[ldv*ncv];
  int *iparam;
  iparam = new int[11];
  iparam[0] = 1;
  iparam[2] = 3*n;           // Maximum number of iterations
  iparam[6] = 1;             // Exact shifting
  int *ipntr;
  ipntr = new int[11];
  double *workd;
  workd = new double[3*n];
  double *workl;
  workl = new double[ncv*(ncv+8)];
  int lworkl = ncv*(ncv+8);
  int info = 0;
  int rvec = 1;
  int *select;
  select = new int[ncv];
  double *d;
  d = new double[2*ncv];
  double sigma;
  int ierr;
  do {
    dsaupd_(&ido, bmat, &n, which, &nev, &tol, resid, &ncv, v, &ldv, iparam, ipntr, workd, workl, &lworkl, &info);
    // Matrix-vector product:
    if ((ido==1)||(ido==-1)) {
      double* in = workd+ipntr[0]-1;
      double* out = workd+ipntr[1]-1;
      for (unsigned int a = 0; a < aGraph.size(); a++) {
        out[a] = 0.0;
        for (unsigned int b = 0; b < aGraph[a].size(); b++) {
          int aNode = aGraph[a][b].mNode;
          out[a] += aGraph[a][b].mWeight*in[aNode];
        }
      }
    }
  } while ((ido==1)||(ido==-1));
  // Get eigenvectors
  char dummy[5] = "All"; 
  dseupd_(&rvec, dummy, select, d, v, &ldv, &sigma, bmat, &n, which, &nev, &tol, resid, &ncv, v, &ldv, iparam, ipntr, workd, workl, &lworkl, &ierr);
  for (int j = 0; j < n; j++)
    X(j) = (v[j]);
  delete resid;
  delete v;
  delete iparam;
  delete ipntr;
  delete workd;
  delete workl;
  delete select;
  delete d;
}

// lanczos ---------------------------------------------------------------------
void lanczos(CGraph& aGraph, CVector<bool>& aValidNodes, CVector<float>& X) {
  // Problem size
  int n = aGraph.size();
  X.setSize(n);
  // Number of eigenvectors
  int nev = 1;
  // Standard eigenvalue problem
  int ido = 0;
  char bmat[2] = "I";
  // Searching the largest eigenvalues
  char which[3] = "LA";
  // Precision
  double tol = 0.0;
  double *resid;
  resid = new double[n];
  // Number of Ritz vectors
  int ncv = 5*nev;
  if (ncv>n) ncv = n;
  double *v;
  int ldv = n;
  v = new double[ldv*ncv];
  int *iparam;
  iparam = new int[11];
  iparam[0] = 1;
  iparam[2] = 3*n;           // Maximum number of iterations
  iparam[6] = 1;             // Exact shifting
  int *ipntr;
  ipntr = new int[11];
  double *workd;
  workd = new double[3*n];
  double *workl;
  workl = new double[ncv*(ncv+8)];
  int lworkl = ncv*(ncv+8);
  int info = 0;
  int rvec = 1;
  int *select;
  select = new int[ncv];
  double *d;
  d = new double[2*ncv];
  double sigma;
  int ierr;
  do {
    dsaupd_(&ido, bmat, &n, which, &nev, &tol, resid, &ncv, v, &ldv, iparam, ipntr, workd, workl, &lworkl, &info);
    // Matrix-vector product:
    if ((ido==1)||(ido==-1)) {
      double* in = workd+ipntr[0]-1;
      double* out = workd+ipntr[1]-1;
      for (unsigned int a = 0; a < aGraph.size(); a++) {
        out[a] = 0.0;
        if (aValidNodes(a) == false) continue;
        for (unsigned int b = 0; b < aGraph[a].size(); b++) {
          int aNode = aGraph[a][b].mNode;
          float w = aGraph[a][b].mWeight;
          if (w > 0 && aValidNodes(aNode)) out[a] += w*in[aNode];
        }
      }
    }
  } while ((ido==1)||(ido==-1));
  // Get eigenvectors
  char dummy[5] = "All";
  dseupd_(&rvec, dummy, select, d, v, &ldv, &sigma, bmat, &n, which, &nev, &tol, resid, &ncv, v, &ldv, iparam, ipntr, workd, workl, &lworkl, &ierr);
  for (int j = 0; j < n; j++)
    X(j) = v[j];
  delete resid;
  delete v;
  delete iparam;
  delete ipntr;
  delete workd;
  delete workl;
  delete select;
  delete d;
}

// ncut ------------------------------------------------------------------------
void ncut(CGraph& aGraph, CVector<bool>& aValidNodes, CMatrix<float>& eigVecs, CVector<float>& eigVals, int aRegionNo) {
  int aValidCount = 0;
  for (unsigned int a = 0; a < aGraph.size(); a++) 
    if (aValidNodes(a)) aValidCount++;
  CVector<int> aNodeToValidNode(aGraph.size());
  CVector<float> D(aValidCount,0);
  int ai = 0;
  for (unsigned int a = 0; a < aGraph.size(); a++) {
    if (aValidNodes(a) == false) continue;
    aNodeToValidNode(a) = ai;
    for (unsigned int b = 0; b < aGraph[a].size(); b++) {
      int aNode = aGraph[a][b].mNode;
      if (aValidNodes(aNode)) D(ai) += fabs(aGraph[a][b].mWeight);
    }
    ai++;
  }
  CVector<float> invD(D.size());
  for (int i = 0; i < D.size(); i++)
    invD(i) = 1.0f/sqrt(D(i)+0.001f);
  // Lanczos method
  int n = aValidCount;
  int nev = 100;
  int ido = 0;
  char bmat[2] = "I";
  char which[3] = "SA";
  double tol = 0.01;
  double* resid = new double[n];
  int ncv = 2*nev;
  if (ncv>n) ncv = n;
  int ldv = n;
  double* v = new double[ldv*ncv];
  int iparam[11];
  iparam[0] = 1;
  iparam[2] = 4*n;           // Maximum number of iterations
  iparam[6] = 1;             // Exact shifting
  int ipntr[11];
  double* workd = new double[3*n];
  double* workl = new double[ncv*(ncv+8)];
  int lworkl = ncv*(ncv+8);
  int info = 0;
  int rvec = 1;
  int* select = new int[ncv];
  double* d = new double[2*ncv];
  double sigma;
  int ierr;
  do {
    dsaupd_(&ido, bmat, &n, which, &nev, &tol, resid, &ncv, v, &ldv, iparam, ipntr, workd, workl, &lworkl, &info);
    // Matrix-vector product:
    if ((ido==1)||(ido==-1)) {
      double* in = workd+ipntr[0]-1;
      double* out = workd+ipntr[1]-1;
      for (unsigned int a = 0; a < aGraph.size(); a++) {
        if (aValidNodes(a) == false) continue;
        int ai = aNodeToValidNode(a);
        out[ai] = in[ai];
        for (unsigned int b = 0; b < aGraph[a].size(); b++) {
          int aNode = aGraph[a][b].mNode;
          if (aValidNodes(aNode) == false) continue;
          aNode = aNodeToValidNode(aNode);
          out[ai] -= aGraph[a][b].mWeight*invD(ai)*invD(aNode)*in[aNode];
        }
      }
    }
  } while ((ido==1)||(ido==-1));
  // Get eigenvectors
  char dummy[5] = "All";
  dseupd_(&rvec, dummy, select, d, v, &ldv, &sigma, bmat, &n, which, &nev, &tol, resid, &ncv, v, &ldv, iparam, ipntr, workd, workl, &lworkl, &ierr);
  delete resid;
  delete workd;
  delete workl;
  delete select;
  // Find number of relevant eigenvectors
  int num = 0;
  for (int j = 1; j < nev; j++)
    if (d[j] < 0.2f) num++;
  if (num < aRegionNo) num = aRegionNo;
  // Invert eigenvalues
  eigVals.setSize(num);
  for (int j = 1; j <= num; j++)
    eigVals(j-1) = 1.0f/sqrt(d[j]);
  // Convert eigenvector of ordinary eigenproblem to eigenvector of general eigenproblem
  eigVecs.setSize(n,num);
  for (int j = 1; j <= num; j++)
    for (int i = 0; i < n; i++)
      eigVecs(i,j-1) = v[j*n+i]*invD(i);
  delete v;
  delete d;
}


