#ifndef COpticFlowH
#define COpticFlowH

#include <CVector.h>
#include <CMatrix.h>
#include <CTensor.h>
#include <CTensor4D.h>

class COpticFlow {
public:
  COpticFlow() {}

  // Variational optical flow methods
  static void hornSchunck(CTensor<float>& aFirst, CTensor<float>& aSecond,
                          CTensor<float>& aFlow,
                          float aSigma = 1.0f, float aAlpha = 500.0f,
                          int aSORIterations = 200, float aOmega = 1.95f);
  static void warpingLinear(CTensor<float>& aFirst, CTensor<float>& aSecond,
                            CTensor<float>& aFlow,
                            float aSigma = 0.9f, float aAlpha = 500.0f, float aGamma = 0.0f,
                            float aEta = 0.9f, int aSORIterations = 20, float aOmega = 1.9f);
  static void warpingLinear(CTensor<float>& aFirst, CTensor<float>& aSecond, CMatrix<float>& aConfidence,
                            CTensor<float>& aFlow,
                            float aSigma = 0.9f, float aAlpha = 500.0f, float aGamma = 0.0f,
                            float aEta = 0.9f, int aSORIterations = 20, float aOmega = 1.9f);
  static void warpingNonlinear(CTensor<float>& aFirst, CTensor<float>& aSecond,
                               CTensor<float>& aFlow,
                               CMatrix<float>* aEdgeMap = 0,
                               float aSigma = 0.7f, float aAlpha = 80.0f, float aGamma = 5.0f, float aEdgeWeight = 0.00025f,
                               float aEta = 0.95f, int aFixedPointIterations = 5, int aSORIterations = 5, float aOmega = 1.85);
  static void warpingGeometric(CTensor<float> aFirst, CTensor<float> aSecond,
                               CTensor<float>& aFlow,
                               CTensor<float>& aGeometricU, CTensor<float>& aGeometricV, CTensor<float>& aGeometricConfidence,
                               CMatrix<float>* aEdgeMap = 0,
                               float aSigma = 0.9f, float aAlpha = 80.0f, float aBeta = 2000.0f, float aGamma = 5.0f, float aEdgeWeight = 0.00025f,
                               float aEta = 0.95f, int aFixedpointIterations = 5, int aSORIterations = 5, float aOmega = 1.85f, float aDelta = 1.0f);

  // Visualization of flow fields as color plots
  static void cartesianToRGB (float x, float y, float& R, float& G, float& B);
  static void flowToImage(CTensor<float>& aFlow, CTensor<float>& aImage, float aDivisor = 1.0f);

  // Others
  static void warp(const CTensor<float>& aImage, CTensor<float>& aWarped, CMatrix<bool>& aOutside, CTensor<float>& aFlow);
  static void writeMiddlebury(CTensor<float>& aFlow, const char* aFilename);
  static bool readMiddlebury(const char* aFilename, CTensor<float>& aFlow, bool aFileNotFoundMessage = true);
  static float angularError(CTensor<float>& aFlow, CTensor<float>& aCorrect);

private:
  static void linearIteration(CTensor<float>& aFirst, CTensor<float>& aSecond, CMatrix<float>& aConfidence, CTensor<float>& aFlow, float aAlpha, float aGamma, int aSORIterations, float aOmega);
  static void diffusivity(CTensor<float>& aFlow, CTensor<float>& aFlowIncrement, CMatrix<float>& aResultX, CMatrix<float>& aResultY, CMatrix<float>* aEdgeMap = 0, float aEdgeWeight = 0.00025f);
  static void diffusionTensor(CTensor<float>& aFlow, CTensor<float>& aFlowIncrement, CMatrix<float>& aOriX, CMatrix<float>& aOriY, CMatrix<float>& A, CMatrix<float>& B, CMatrix<float>& C);
  static void nonlinearIteration(CTensor<float>& aFirst, CTensor<float>& aSecond, CMatrix<float>& aConfidence, CTensor<float>& aFlow, CMatrix<float>* aEdgeMap, float aAlpha, float aGamma, float aEdgeWeight, int aFixedPointIterations, int aSORIterations, float aOmega);
  static void nonlinearIteration(CTensor<float>& aFirst, CTensor<float>& aSecond, CMatrix<float>& aConfidence, CTensor<float>& aGeometricU, CTensor<float>& aGeometricV, CTensor<float>& aGeometricConfidence, CTensor<float>& aFlow, CMatrix<float>* aEdgeMap, float aAlpha, float aBeta, float aGamma, float aEdgeWeight, int aFixedPointIterations, int aSORIterations, float aOmega, float aDelta);
  static void nonlinearAnisotropic(CTensor<float>& aFirst, CTensor<float>& aSecond, CMatrix<float>& aConfidence, CTensor<float>& aGeometricU, CTensor<float>& aGeometricV, CTensor<float>& aGeometricConfidence, CTensor<float>& aFlow, CMatrix<float>* aEdgeMap, float aAlpha, float aBeta, float aGamma, float aEdgeWeight, int aFixedPointIterations, int aSORIterations, float aOmega);
};

#endif

