#include <string>
#include <vector>
#include <algorithm>
#include <ctime>
#include <sys/resource.h>
#include "CTensor.h"
#include "CFilter.h"
#include "CDiffusion.h"
#include "CMatching.h"
#include "COpticFlow.h"
#include "CGraph.h"
#include "CKMeans.h"

//#define SHOWEIGENVECTORS
//#define SHOWAFTERKMEANS
//#define SHOWAFTERMERGING

CVector<std::string> mInput;
std::string mFilename;
std::string mInputDir;
std::string mResultDir;
std::vector<CTrack> mTracks;
int mStartFrame;
int mStep;
int mRegionNo;
int mSequenceLength;
CVector<float> mEigVals;
int mXSize,mYSize;
CGraph mGraph;
std::queue<int> mSplitQueue;

CMatrix<float> mColorCode;
std::ofstream mStatus;

// buildColorCode
void buildColorCode() {
  mColorCode.setSize(3,15000);
  for (int i = 0; i < 256; i++) {
    mColorCode(0,i) = 0;
    mColorCode(1,i) = i;
    mColorCode(2,i) = 255;
  }
  for (int i = 0; i < 256; i++) {
    mColorCode(0,i+256) = 0;
    mColorCode(1,i+256) = 255;
    mColorCode(2,i+256) = 255-i;
  }
  for (int i = 0; i < 256; i++) {
    mColorCode(0,i+512) = i;
    mColorCode(1,i+512) = 255;
    mColorCode(2,i+512) = 0;
  }
  for (int i = 0; i < 256; i++) {
    mColorCode(0,i+768) = 255;
    mColorCode(1,i+768) = 255-i;
    mColorCode(2,i+768) = 0;
  }
  for (int i = 0; i < 256; i++) {
    mColorCode(0,i+1024) = 255;
    mColorCode(1,i+1024) = 0;
    mColorCode(2,i+1024) = i;
  }
  for (int i = 1280; i < mColorCode.ySize(); i++) {
    mColorCode(0,i) = 255;
    mColorCode(1,i) = 0;
    mColorCode(2,i) = 255;
  }
}

bool readTrackFromFile(std::vector<CTrack>& aTracks) {
  std::string aFilename = mResultDir+mFilename;
  char buffer[100];
  sprintf(buffer,"Tracks_%03d_%03d_%02d.txt",mStartFrame,mSequenceLength,mStep);
  std::ifstream aIn((aFilename+buffer).c_str());
  if (!aIn.is_open()) return false;
  int aSize;
  aIn >> aSize;
  aTracks.clear();
  for (int i = 0; i < aSize; i++) {
    aTracks.push_back(CTrack());
    aTracks.back().readFromFile(aIn);
  }
  return true;
}

void writeTrackToFile(std::vector<CTrack>& aTracks) {
  std::string aFilename = mResultDir+mFilename;
  char buffer[100];
  sprintf(buffer,"Tracks_%03d_%03d_%02d.txt",mStartFrame,mSequenceLength,mStep);
  std::ofstream aOut((aFilename+buffer).c_str());
  aOut << aTracks.size() << std::endl;
  for (unsigned int i = 0; i < aTracks.size(); i++)
    aTracks[i].writeToFile(aOut);
}

void readGraphFromFile(CGraph& aGraph) {
  std::string aFilename = mResultDir+mFilename;
  char buffer[100];
  sprintf(buffer,"Graph_%03d_%03d_%02d.bin",mStartFrame,mSequenceLength,mStep);
  std::ifstream aIn((aFilename+buffer).c_str(),std::ios::binary);
  aGraph.clear();
  int aSize;
  aIn.read((char*)&aSize,sizeof(aSize));
  for (unsigned int a = 0; a < aSize; a++) {
    aGraph.push_back(std::vector<CEdge>());
    int aSize2;
    aIn.read((char*)&aSize2,sizeof(aSize2));
    for (int b = 0; b < aSize2; b++) {
      aGraph[a].push_back(CEdge());
      aGraph[a].back().readFromFile(aIn);
    }
  }
  int aSize3;
  aIn.read((char*)&mRegionNo,sizeof(mRegionNo));
  aIn.read((char*)&aSize3,sizeof(aSize3));
  mEigVals.setSize(aSize3);
  aIn.read((char*)mEigVals.data(),sizeof(float)*mEigVals.size());
}

void writeGraphToFile(CGraph& aGraph) {
  std::string aFilename = mResultDir+mFilename;
  char buffer[100];
  sprintf(buffer,"Graph_%03d_%03d_%02d.bin",mStartFrame,mSequenceLength,mStep);
  std::ofstream aOut((aFilename+buffer).c_str(),std::ios::binary);
  int dummy = (int)aGraph.size();
  aOut.write((char*)&dummy,sizeof(dummy));
  for (unsigned int a = 0; a < aGraph.size(); a++) {
    dummy = (int)aGraph[a].size();
    aOut.write((char*)&dummy,sizeof(dummy));
    for (unsigned int b = 0; b < aGraph[a].size(); b++)
      aGraph[a][b].writeToFile(aOut);
  }
  aOut.write((char*)&mRegionNo,sizeof(mRegionNo));
  dummy = mEigVals.size();
  aOut.write((char*)&dummy,sizeof(dummy));
  aOut.write((char*)mEigVals.data(),sizeof(float)*mEigVals.size());
}

class CPoint {
public:
  CPoint() {}
  float x,y,frame;
};

class CSimpleTrack {
public:
  CSimpleTrack() {}
  int mLabel;
  CVector<CPoint> mPoints;
};

// writeTracks
void writeTracks() {
  // Create a filename and open a file
  char buffer[50];
  sprintf(buffer,"Tracks%d.dat",mSequenceLength);
  std::ofstream aFile((mResultDir+buffer).c_str());
  // Write number of frames considered
  aFile << mSequenceLength << std::endl;
  // Write number of tracks
  int aCount = 0;
  for (int i = 0; i < mTracks.size(); i++)
    if (mTracks[i].mLabel >= 0) aCount++;
  aFile << aCount << std::endl;
  // Write each track
  for (int i = 0; i < mTracks.size(); i++) {
    // Ignore tracks marked as outliers
    if (mTracks[i].mLabel < 0) continue;
    // Write label and length of track
    int aSize = mTracks[i].mx.size();
    aFile << mTracks[i].mLabel << " " << aSize+1 << std::endl;
    // Write x,y coordinates and the frame number when the track was set up
    aFile << mTracks[i].mox << " " << mTracks[i].moy << " " << mTracks[i].mSetupTime << std::endl;
    // Write x,y coordinates and frame number of the tracked point in the successive frames
    for (int j = 0; j < aSize; j++)
      aFile << mTracks[i].mx[j] << " " << mTracks[i].my[j] << " " << mTracks[i].mSetupTime+j+1 << std::endl;
  }
}

class CSizeStruct {
public:
  CSizeStruct(int aSize, int aLabel) { mSize = aSize; mLabel = aLabel; }
  int mSize,mLabel;
};

bool operator<(const CSizeStruct& a, const CSizeStruct& b) {
  return a.mSize > b.mSize;
}

float computeError(std::vector<CTrack>& aTracks) {
  // Read ground truth file if there is one
  std::ifstream aTruth((mInputDir+mFilename+"_truth.txt").c_str());
  if (aTruth.is_open() == false) return -1.0f;
  int aPointCount;
  unsigned int aSequenceLength;
  int aShift;
  aTruth >> aPointCount >> aSequenceLength >> aShift;
  CMatrix<int> X(aPointCount,aSequenceLength);
  CMatrix<int> Y(aPointCount,aSequenceLength);
  CMatrix<int> aLabel(aPointCount,aSequenceLength);
  std::vector<CSizeStruct> aSizes;
  for (unsigned int j = 0; j < aSequenceLength; j++)
    for (int i = 0; i < aPointCount; i++) {
      aTruth >> X(i,j) >> Y(i,j) >> aLabel(i,j);
      while (aLabel(i,j) > aSizes.size())
        aSizes.push_back(CSizeStruct(0,aSizes.size()+1));
      aSizes[aLabel(i,j)-1].mSize++;
    }
  std::sort(aSizes.begin(),aSizes.end());
  aSequenceLength = mSequenceLength;
  // For each ground truth point, check for closest computed point
  CVector<bool> aAllowedLabels(mRegionNo,true);
  int aPositives = 0;
  int aNegatives = 0;
  CTensor<float>* aShow = new CTensor<float>[aSequenceLength+aShift];
  for (unsigned int t = 0; t < aSequenceLength+aShift; t++)
    aShow[t].readFromPPM(mInput(t).c_str());
  for (unsigned int r = 0; r < aSizes.size(); r++) {
    CVector<int> aAssigned(mRegionNo,0);
    for (int p = 0; p < aPointCount; p++) {
      if (aLabel(p,0) != aSizes[r].mLabel) continue;
      for (unsigned int t = aShift; t < aSequenceLength+aShift; t++) {
        int x1 = X(p,t-aShift)-1; if (x1 < 0) x1 = 0;
        int y1 = Y(p,t-aShift)-1; if (y1 < 0) y1 = 0;
        int x2 = X(p,t-aShift)+1; if (x2 >= mXSize) x2 = mXSize-1;
        int y2 = Y(p,t-aShift)+1; if (y2 >= mYSize) y2 = mYSize-1;
        for (int y = y1; y <= y2; y++)
          for (int x = x1; x <= x2; x++) {
            aShow[t](x,y,0) = mColorCode(0,400*r);
            aShow[t](x,y,1) = mColorCode(1,400*r);
            aShow[t](x,y,2) = mColorCode(2,400*r);
          }
        float bestdist = 1e9f;
        char bestLabel = 0;
        for (unsigned int i = 0; i < aTracks.size(); i++) {
          if (aTracks[i].mLabel < 0) continue;
          unsigned int aStart = aTracks[i].mSetupTime;
          if (aStart > t) continue;
          if (aStart+aTracks[i].mx.size() < t) continue;
          float ax,ay;
          if (aStart == t) {
            ax = aTracks[i].mox; ay = aTracks[i].moy;
          }
          else {
            ax = aTracks[i].mx[t-aStart-1]; ay = aTracks[i].my[t-aStart-1];
          }
          float temp = ax-X(p,t-aShift);
          float dist = temp*temp;
          temp = ay-Y(p,t-aShift);
          dist += temp*temp;
          if (dist < bestdist) {
            bestdist = dist;
            bestLabel = aTracks[i].mLabel;
          }
        }
        aAssigned(bestLabel)++;
      }
    }
    int max = -1;
    int bestCount = 0;
    for (int i = 0; i < mRegionNo; i++)
      if (aAllowedLabels(i) && aAssigned(i) > bestCount) {
        bestCount = aAssigned(i);
        max = i;
      }
    if (max >= 0) {
      aPositives += aAssigned(max);
      aNegatives += aAssigned.sum()-aAssigned(max);
      aAllowedLabels(max) = false;
    }
    else aNegatives += aAssigned.sum();
    for (int p = 0; p < aPointCount; p++) {
      if (aLabel(p,0) != aSizes[r].mLabel) continue;
      for (unsigned int t = aShift; t < aSequenceLength+aShift; t++) {
        float bestdist = 1e9f;
        char bestLabel = 0;
        for (unsigned int i = 0; i < aTracks.size(); i++) {
          if (aTracks[i].mLabel < 0) continue;
          unsigned int aStart = aTracks[i].mSetupTime;
          if (aStart > t) continue;
          if (aStart+aTracks[i].mx.size() < t) continue;
          float ax,ay;
          if (aStart == t) {
            ax = aTracks[i].mox; ay = aTracks[i].moy;
          }
          else {
            ax = aTracks[i].mx[t-aStart-1]; ay = aTracks[i].my[t-aStart-1];
          }
          float temp = ax-X(p,t-aShift);
          float dist = temp*temp;
          temp = ay-Y(p,t-aShift);
          dist += temp*temp;
          if (dist < bestdist) {
            bestdist = dist;
            bestLabel = aTracks[i].mLabel;
          }
        }
        if (bestLabel != max) {
          int x1 = X(p,t-aShift)-3; if (x1 < 0) x1 = 0;
          int y1 = Y(p,t-aShift)-3; if (y1 < 0) y1 = 0;
          int x2 = X(p,t-aShift)+3; if (x2 >= mXSize) x2 = mXSize-1;
          int y2 = Y(p,t-aShift)+3; if (y2 >= mYSize) y2 = mYSize-1;
          for (int y = y1; y <= y2; y++)
            for (int x = x1; x <= x2; x++) {
              aShow[t](x,y,0) = mColorCode(0,400*r);
              aShow[t](x,y,1) = mColorCode(1,400*r);
              aShow[t](x,y,2) = mColorCode(2,400*r);
            }
        }
      }
    }
  }
  for (unsigned int t = aShift; t < aSequenceLength+aShift; t++) {
    char buffer[200];
    sprintf(buffer,"%s%03d.ppm",(mResultDir+"Truth").c_str(),t);
    aShow[t].writeToPPM(buffer);
  }
  float aResult = 100.0f*aNegatives/(float)(aPositives+aNegatives);
  std::ofstream aOut((mResultDir+"error.txt").c_str());
  aOut << aPositives << " " << aNegatives << " " << aResult << std::endl;
  delete[] aShow;
  return aResult;
}

void showSegments(std::vector<CTrack>& aTracks, unsigned int t, int k = -1, bool aDownsample = false) {
  if (mSequenceLength <= (int)t) return;
  int R[16] = {255, 10,200,  0,255,  0,  0,180,255,180,  0,170,100,240,  0,255};
  int G[16] = { 10, 10,190,180,  0,255,255,  0,210,  0,  0,170,100,100,100,140};
  int B[16] = { 10,255,  0,  0,255,255,  0,  0,  0,180,190,170,100, 10,255,  0};
  CTensor<float> aShow(mXSize,mYSize,3,255);
  for (unsigned int i = 0; i < aTracks.size(); i++) {
    int aTime = t-aTracks[i].mSetupTime;
    if (aTime < 0) continue;
    if ((int)aTracks[i].mx.size() < aTime) continue;
    int aLabel = aTracks[i].mLabel;
    int x,y;
    if (aTime > 0) {
      x = (int)aTracks[i].mx[aTime-1];
      y = (int)aTracks[i].my[aTime-1];
    }
    else {
      x = aTracks[i].mox;
      y = aTracks[i].moy;
    }
    int r = mStep/4;
    int x1 = x-r; if (x1 < 0) x1 = 0;
    int y1 = y-r; if (y1 < 0) y1 = 0;
    int x2 = x+r; if (x2 >= aShow.xSize()) x2 = aShow.xSize()-1;
    int y2 = y+r; if (y2 >= aShow.ySize()) y2 = aShow.ySize()-1;
    for (int ay = y1; ay <= y2; ay++)
      for (int ax = x1; ax <= x2; ax++)
        if (aLabel < 0) {
          aShow(ax,ay,0) = 230;
          aShow(ax,ay,1) = 230;
          aShow(ax,ay,2) = 230;
        }
        else {
          aShow(ax,ay,0) = R[aLabel];
          aShow(ax,ay,1) = G[aLabel];
          aShow(ax,ay,2) = B[aLabel];
        }
  }
  char buffer[200];
  if (k > 0) sprintf(buffer,"Segments_%03d_%03d.ppm",k,t+mStartFrame);
  else sprintf(buffer,"Segments%03d.ppm",t+mStartFrame);
  if (aDownsample) aShow.downsample(aShow.xSize()/2,aShow.ySize()/2);
  aShow.writeToPPM((mResultDir+buffer).c_str());
}

void showEigenvectors(std::vector<CTrack>& aTracks, unsigned int t) {
  if (mSequenceLength <= (int)t) return;
  int aDim = mEigVals.size();
  if (aDim > 16) aDim = 16;
  CMatrix<float>* aShow = new CMatrix<float>[aDim];
  for (int i = 0; i < aDim; i++) {
    aShow[i].setSize(mXSize,mYSize);
    aShow[i] = 0;
  }
  for (unsigned int i = 0; i < aTracks.size(); i++) {
    int aTime = t-aTracks[i].mSetupTime;
    if (aTime < 0) continue;
    if ((int)aTracks[i].mx.size() <= aTime) continue;
    if (aTracks[i].mProjection.size() < aDim) continue;
    int x,y;
    if (aTime > 0) {
      x = (int)aTracks[i].mx[aTime-1];
      y = (int)aTracks[i].my[aTime-1];
    }
    else {
      x = aTracks[i].mox;
      y = aTracks[i].moy;
    }
    int r = mStep/4;
    int x1 = x-r; if (x1 < 0) x1 = 0;
    int y1 = y-r; if (y1 < 0) y1 = 0;
    int x2 = x+r; if (x2 >= aShow[0].xSize()) x2 = aShow[0].xSize()-1;
    int y2 = y+r; if (y2 >= aShow[0].ySize()) y2 = aShow[0].ySize()-1;
    for (int k = 0; k < aDim; k++)
      for (int ay = y1; ay <= y2; ay++)
        for (int ax = x1; ax <= x2; ax++) {
          if (aTracks[i].mProjection(k) < -1) aShow[k](ax,ay) = 0;
          else {
            aShow[k](ax,ay) = 60+195*aTracks[i].mProjection(k);
            if (aShow[k](ax,ay) > 255) aShow[k](ax,ay) = 255;
          }
        }
  }
  char buffer[200];
  for (int k = 0; k < aDim; k++) {
    sprintf(buffer,"EigVecs%03d_%02d.pgm",t+mStartFrame,k);
    aShow[k].downsample(mXSize/2,mYSize/2);
    aShow[k].writeToPGM((mResultDir+buffer).c_str());
  }
  delete[] aShow;
}

void showFlow(CTensor<float>& aForwardFlow, CTensor<float>& aBackwardFlow, int t) {
  char buffer[200];
  CTensor<float> aShow;
  sprintf(buffer,"ForwardFlow%03d.ppm",t+mStartFrame);
  COpticFlow::flowToImage(aForwardFlow,aShow,4);
  aShow.writeToPPM((mResultDir+buffer).c_str());
  sprintf(buffer,"BackwardFlow%03d.ppm",t+mStartFrame);
  COpticFlow::flowToImage(aBackwardFlow,aShow,4);
  aShow.writeToPPM((mResultDir+buffer).c_str());
}

void showWork(CMatrix<float>& aWork, CVector<int>& aNodeToIndex, std::vector<CTrack>& aTracks,float aFactor) {
  CMatrix<float>* aShow = new CMatrix<float>[aWork.ySize()];
  int aDim = aWork.ySize();
  for (int i = 0; i < aDim; i++) {
    aShow[i].setSize(mXSize,mYSize);
    aShow[i] = 0;
  }
  for (int ai = 0; ai < aWork.xSize(); ai++) {
    int i = aNodeToIndex(ai);
    int aTime = 0-aTracks[i].mSetupTime;
    if (aTime < 0) continue;
    if ((int)aTracks[i].mx.size() <= aTime) continue;
    int x,y;
    if (aTime > 0) {
      x = (int)aTracks[i].mx[aTime-1];
      y = (int)aTracks[i].my[aTime-1];
    }
    else {
      x = aTracks[i].mox;
      y = aTracks[i].moy;
    }
    int r = mStep/4;
    int x1 = x-r; if (x1 < 0) x1 = 0;
    int y1 = y-r; if (y1 < 0) y1 = 0;
    int x2 = x+r; if (x2 >= aShow[0].xSize()) x2 = aShow[0].xSize()-1;
    int y2 = y+r; if (y2 >= aShow[0].ySize()) y2 = aShow[0].ySize()-1;
    for (int k = 0; k < aDim; k++)
      for (int ay = y1; ay <= y2; ay++)
        for (int ax = x1; ax <= x2; ax++)
          aShow[k](ax,ay) = 50+aFactor*155.0f*aWork(ai,k);
  }
  char buffer[200];
  for (int k = 0; k < aDim; k++) {
    sprintf(buffer,"%s%d.pgm",(mResultDir+"Work").c_str(),k);
    aShow[k].downsample(mXSize/2,mYSize/2);
  }
  delete[] aShow;
}

// For euclideanDistanceTransform
void dt(CVector<float>& f, CVector<float>& d, int n) {
  d.setSize(n);
  int *v = new int[n];
  float *z = new float[n+1];
  int k = 0;
  v[0] = 0;
  z[0] = -10e20;
  z[1] = 10e20;
  for (int q = 1; q <= n-1; q++) {
    float s  = ((f[q]+q*q)-(f(v[k])+v[k]*v[k]))/(2*(q-v[k]));
    while (s <= z[k]) {
      k--;
      s  = ((f[q]+q*q)-(f(v[k])+v[k]*v[k]))/(2*(q-v[k]));
    }
    k++;
    v[k] = q;
    z[k] = s;
    z[k+1] = 10e20;
  }
  k = 0;
  for (int q = 0; q <= n-1; q++) {
    while (z[k+1] < q)
      k++;
    int help = q-v[k];
    d(q) = help*help + f(v[k]);
  }
  delete[] v;
  delete[] z;
}

// euclideanDistanceTransform
void euclideanDistanceTransform(CMatrix<float>& aMatrix) {
  int aXSize = aMatrix.xSize();
  int aYSize = aMatrix.ySize();
  CVector<float> f(NMath::max(aXSize,aYSize));
  // Transform along columns
  for (int x = 0; x < aXSize; x++) {
    for (int y = 0; y < aYSize; y++)
      f(y) = aMatrix(x,y);
    CVector<float> d;
    dt(f,d,aYSize);
    for (int y = 0; y < aYSize; y++)
      aMatrix(x,y) = d(y);
  }
  // Transform along rows
  for (int y = 0; y < aYSize; y++) {
    int aOffset = y*aXSize;
    for (int x = 0; x < aXSize; x++)
      f(x) = aMatrix.data()[x+aOffset];
    CVector<float> d;
    dt(f,d,aXSize);
    for (int x = 0; x < aXSize; x++)
      aMatrix.data()[x+aOffset] = d(x);
  }
}

void computeLocalFlowVariation(CTensor<float>& aFlow, CMatrix<float>& aFlowVar) {
  int aXSize = aFlow.xSize();
  int aYSize = aFlow.ySize();
  int aSize = aFlow.size();
  CTensor<float> aStats(aXSize,aYSize,4);
  aStats.paste(aFlow,0,0,0);
  for (int i = 0; i < aSize; i++)
    aStats.data()[i+aSize] = aFlow.data()[i]*aFlow.data()[i];
  aXSize = 1+(aXSize >> 1);
  aYSize = 1+(aYSize >> 1);
  aStats.downsample(aXSize,aYSize);
  CDiffusion aDiffusion;
  aDiffusion.mP = -1.0f;
  CTensor<float> aFlowCoarse(aFlow);
  aFlowCoarse.downsample(aXSize,aYSize);
  CMatrix<float> GX(aXSize,aYSize,1);
  CMatrix<float> GY(aXSize,aYSize,1);
  aDiffusion.isotropicDiffMatrixMultiChannel(aFlowCoarse,GX,GY);
  aDiffusion.mTau2D = 0.5f;
  int aIterations = (int)ceil(10.0f/aDiffusion.mTau2D);
  CMatrix<float> aTemp(aXSize,aYSize);
  for (int k = 0; k < aIterations; k++)
    for (int z = 0; z < aStats.zSize(); z++) {
      aStats.getMatrix(aTemp,z);
      aDiffusion.isotropicDiffuseSemiImplicit(aTemp,GX,GY);
      aStats.putMatrix(aTemp,z);
    }
  aSize = aXSize*aYSize;
  int a2Size = 2*aSize;
  int a3Size = 3*aSize;
  aFlowVar.setSize(aXSize,aYSize);
  for (int i = 0; i < aSize; i++) {
    float help = aStats.data()[i+a2Size]-aStats.data()[i]*aStats.data()[i]+aStats.data()[i+a3Size]-aStats.data()[i+aSize]*aStats.data()[i+aSize];
    if (help < 0.0f) help = 0.0f;
    aFlowVar.data()[i] = sqrt(help);
  }
  aFlowVar.upsample(mXSize,mYSize);
}

void computeNearestNeighbors(CTrack& A, CTrack& B, int aIndexA, int aIndexB) {
  // Find common frames of the two tracks
  int aStartTime = NMath::max(A.mSetupTime,B.mSetupTime);
  int aLength = NMath::min((int)A.mx.size()+A.mSetupTime-aStartTime,(int)B.mx.size()+B.mSetupTime-aStartTime);
  if (aLength <= 0) return;
  // Coordinates of first common point
  int aTimeA = aStartTime-A.mSetupTime-1;
  int aTimeB = aStartTime-B.mSetupTime-1;
  int cx1,cy1,cx2,cy2;
  if (aTimeA < 0) {cx1 = A.mox; cy1 = A.moy;}
  else {cx1 = (int)A.mx[aTimeA]; cy1 = (int)A.my[aTimeA];}
  if (aTimeB < 0) {cx2 = B.mox; cy2 = B.moy;}
  else {cx2 = (int)B.mx[aTimeB]; cy2 = (int)B.my[aTimeB];}
  // Average spatial position of the tracked points over common length of track
  float x1 = cx1; float y1 = cy1;
  float x2 = cx2; float y2 = cy2;
  for (int t = aStartTime; t < aStartTime+aLength; t++) {
    float ax = A.mx[t-A.mSetupTime];
    float ay = A.my[t-A.mSetupTime];
    float bx = B.mx[t-B.mSetupTime];
    float by = B.my[t-B.mSetupTime];
    x1 += ax; y1 += ay;
    x2 += bx; y2 += by;
  }
  float invLength = 1.0f/(aLength+1);
  x1 *= invLength; x2 *= invLength;
  y1 *= invLength; y2 *= invLength;
  float temp = x1-x2;
  float spdist = temp*temp;
  temp = y1-y2;
  spdist += temp*temp;
  // Save spatially nearest neighbors
  bool alreadyThere = false;
  for (int s = 0; s < A.mNearestNeighbors.size(); s++)
    if (A.mNearestNeighbors(s) == aIndexB) {alreadyThere = true; break;}
  if (alreadyThere == false) {
    int k = A.mNearestNeighbors.size();
    while (k > 0 && A.mNearestNeighborsDX(k-1)*A.mNearestNeighborsDX(k-1)+A.mNearestNeighborsDY(k-1)*A.mNearestNeighborsDY(k-1) >= spdist)
      k--;
    if (k < A.mNearestNeighbors.size()) {
      for (int k2 = A.mNearestNeighbors.size()-1; k2 > k; k2--) {
        A.mNearestNeighborsDX(k2) = A.mNearestNeighborsDX(k2-1);
        A.mNearestNeighborsDY(k2) = A.mNearestNeighborsDY(k2-1);
        A.mNearestNeighbors(k2) = A.mNearestNeighbors(k2-1);
      }
      A.mNearestNeighborsDX(k) = x2-x1;
      A.mNearestNeighborsDY(k) = y2-y1;
      A.mNearestNeighbors(k) = aIndexB;
    }
  }
  alreadyThere = false;
  for (int s = 0; s < B.mNearestNeighbors.size(); s++)
    if (B.mNearestNeighbors(s) == aIndexA) {alreadyThere = true; break;}
  if (alreadyThere == false) {
    int k = B.mNearestNeighbors.size();
    while (k > 0 && B.mNearestNeighborsDX(k-1)*B.mNearestNeighborsDX(k-1)+B.mNearestNeighborsDY(k-1)*B.mNearestNeighborsDY(k-1) >= spdist)
      k--;
    if (k < B.mNearestNeighbors.size()) {
      for (int k2 = B.mNearestNeighbors.size()-1; k2 > k; k2--) {
        B.mNearestNeighborsDX(k2) = B.mNearestNeighborsDX(k2-1);
        B.mNearestNeighborsDY(k2) = B.mNearestNeighborsDY(k2-1);
        B.mNearestNeighbors(k2) = B.mNearestNeighbors(k2-1);
      }
      B.mNearestNeighborsDX(k) = x1-x2;
      B.mNearestNeighborsDY(k) = y1-y2;
      B.mNearestNeighbors(k) = aIndexA;
    }
  }
}

float computeDistance(CTrack& A, CTrack& B, int aIndexA, int aIndexB, float aRank) {
  // Find common frames of the two tracks
  int aStartTime = NMath::max(A.mSetupTime,B.mSetupTime);
  int aLength = NMath::min((int)A.mx.size()+A.mSetupTime-aStartTime,(int)B.mx.size()+B.mSetupTime-aStartTime);
  if (aLength <= 0) return 0.0f;
  // Coordinates of first common point
  int aTimeA = aStartTime-A.mSetupTime-1;
  int aTimeB = aStartTime-B.mSetupTime-1;
  int cx1,cy1,cx2,cy2;
  if (aTimeA < 0) {cx1 = A.mox; cy1 = A.moy;}
  else {cx1 = (int)A.mx[aTimeA]; cy1 = (int)A.my[aTimeA];}
  if (aTimeB < 0) {cx2 = B.mox; cy2 = B.moy;}
  else {cx2 = (int)B.mx[aTimeB]; cy2 = (int)B.my[aTimeB];}
  // 1. Average spatial distance of the tracked points over common length of track
  float x1 = cx1; float y1 = cy1;
  float x2 = cx2; float y2 = cy2;
  for (int t = aStartTime; t < aStartTime+aLength; t++) {
    float ax = A.mx[t-A.mSetupTime];
    float ay = A.my[t-A.mSetupTime];
    float bx = B.mx[t-B.mSetupTime];
    float by = B.my[t-B.mSetupTime];
    x1 += ax; y1 += ay;
    x2 += bx; y2 += by;
  }
  float invLength = 1.0f/(aLength+1);
  x1 *= invLength; x2 *= invLength;
  y1 *= invLength; y2 *= invLength;
  float temp = x1-x2;
  float spdist = temp*temp;
  temp = y1-y2;
  spdist += temp*temp;
  // 2. Short term motion analysis: maximum difference in r-frame motion
  float maxSTdist = 0.0f;
  int r = 5;
  if (aLength < r) r = aLength;
  for (int t1 = aStartTime; t1 <= aStartTime+aLength-r; t1++) {
    float ax,ay,bx,by;
    int ta = t1-A.mSetupTime-1;
    int tb = t1-B.mSetupTime-1;
    if (ta < 0) { ax = A.mox; ay = A.moy; }
    else { ax = A.mx[ta]; ay = A.my[ta]; }
    if (tb < 0) { bx = B.mox; by = B.moy; }
    else { bx = B.mx[tb]; by = B.my[tb]; }
    float temp = A.mx[ta+r]-ax-(B.mx[tb+r]-bx);
    float dist = temp*temp;
    temp = A.my[ta+r]-ay-(B.my[tb+r]-by);
    dist += temp*temp;
    float flowvarA;
    if (ta < 0) flowvarA = A.mFlowVar[ta+r];
    else flowvarA = A.mFlowVar[ta+r]-A.mFlowVar[ta];
    float flowvarB;
    if (tb < 0) flowvarB = B.mFlowVar[tb+r];
    else flowvarB = B.mFlowVar[tb+r]-B.mFlowVar[tb];
    float flowvar = NMath::min(flowvarA,flowvarB);
    dist *= 1.0f/(flowvar*flowvar);
    if (dist > maxSTdist) maxSTdist = dist;
  }
  maxSTdist *= 0.1f*sqrt(spdist)/r;
  // 3. Check whether the pair is a nearest neighbor pair
  bool aPositive = false;
  for (int i = 0; i < A.mNearestNeighbors.size(); i++)
    if (aIndexB == A.mNearestNeighbors(i)) { aPositive = true; break; }
  for (int i = 0; i < B.mNearestNeighbors.size(); i++)
    if (aIndexA == B.mNearestNeighbors(i)) { aPositive = true; break; }
  // Final affinities
  float aWeight = exp(-maxSTdist);
  return aWeight;
}

class CTrackItem {
public:
  CTrackItem(int aIndex, float aScore) { mIndex = aIndex; mScore = aScore; }
  int mIndex;
  float mScore;
};

bool operator<(const CTrackItem& a, const CTrackItem& b) {
  return a.mScore > b.mScore;
}

// buildGraph
void buildGraph(std::vector<CTrack>& aTracks, CVector<bool>& aValidTracks, CGraph& aEdges) {
  int aEdgeCount = 0;
  aEdges.clear();
  for (unsigned int i = 0; i < aTracks.size(); i++)
    aEdges.push_back(std::vector<CEdge>());
  // Determine nearest spatial neighbors of each track
  for (unsigned int i = 0; i < aTracks.size(); i++) {
    if (aValidTracks(i) == false) continue;
    for (unsigned int j = i+1; j < aTracks.size(); j++)
      if (aValidTracks(j)) computeNearestNeighbors(aTracks[i],aTracks[j],i,j);
  }
  // Score each track by its length
  int aTotalLength = 0;
  int aCoveredLength = 0;
  std::vector<CTrackItem> aList;
  std::vector<int> aAssigned;
  for (unsigned int i = 0; i < aTracks.size(); i++) {
    if (aValidTracks(i) == false) continue;
    float flowvar = aTracks[i].mFlowVar.back();
    aList.push_back(CTrackItem(i,flowvar));
    aTotalLength += aTracks[i].mx.size();
  }
  float invCount = 1.0f/aList.size();
  // Area covered by tracks so far (comprises distance to closest point)
  CMatrix<float>* aCovered = new CMatrix<float>[mSequenceLength+1];
  float invStep = 1.0f/mStep;
  for (int t = 0; t <= mSequenceLength; t++) {
    aCovered[t].setSize(1+mXSize/mStep,1+mYSize/mStep);
    aCovered[t] = 1e20f;
  }
  do {
    std::sort(aList.begin(),aList.end());
    unsigned int aStop = aList.size()>>2;
    if (aCoveredLength > 0.95*aTotalLength) aStop = aList.size();
    // Add edges...
    for (unsigned int k1 = 0; k1 < aStop; k1++) {
      int i = aList[k1].mIndex;
      // ...between all just accepted tracks...
      for (unsigned int k2 = k1+1; k2 < aStop; k2++) {
        int j = aList[k2].mIndex;
        float w = computeDistance(aTracks[i],aTracks[j],i,j,(aAssigned.size()+k1)*invCount);
        if (w > 0.05 || w < -0.05) {
          aEdges[i].push_back(CEdge(j,w));
          aEdges[j].push_back(CEdge(i,w));
          aTracks[i].mLabel = 0;
          aTracks[j].mLabel = 0;
          aEdgeCount += 2;
        }
      }
      // ...and to previously accepted tracks
      for (unsigned int k2 = 0; k2 < aAssigned.size(); k2++) {
        int j = aAssigned[k2];
        float w = computeDistance(aTracks[i],aTracks[j],i,j,(aAssigned.size()+k1)*invCount);
        if (w > 0.05 || w < -0.05) {
          aEdges[i].push_back(CEdge(j,w));
          aEdges[j].push_back(CEdge(i,w));
          aTracks[i].mLabel = 0;
          aTracks[j].mLabel = 0;
          aEdgeCount += 2;
        }
      }
    }
    // Mark covered areas
    for (unsigned int k1 = 0; k1 < aStop; k1++) {
      int i = aList[k1].mIndex;
      aAssigned.push_back(i);
      aCoveredLength += aTracks[i].mx.size();
      for (unsigned int t = 0; t < aTracks[i].mx.size(); t++) {
        int x = (int)(invStep*aTracks[i].mx[t]);
        int y = (int)(invStep*aTracks[i].my[t]);
        aCovered[t+aTracks[i].mSetupTime](x,y) = 0.0f;
      }
    }
    for (int t = 0; t < mSequenceLength; t++)
      euclideanDistanceTransform(aCovered[t]);
    // Delete assigned tracks from list
    for (unsigned int k1 = 0; k1 < aStop && k1 < aList.size(); k1++) {
      aList[k1] = aList.back();
      aList.pop_back();
    }
    // Recompute scores
    for (unsigned int k1 = 0; k1 < aList.size(); k1++) {
      int i = aList[k1].mIndex;
      float aDistanceCounter = 0.0f;
      for (unsigned int t = 0; t < aTracks[i].mx.size(); t++) {
        int x = (int)(invStep*aTracks[i].mx[t]);
        int y = (int)(invStep*aTracks[i].my[t]);
        aDistanceCounter += aTracks[i].mFlowVar.back()*aCovered[t+aTracks[i].mSetupTime](x,y);
      }
      aList[k1].mScore = aDistanceCounter;
    }
  } while (aCoveredLength < 0.95*aTotalLength);
  delete[] aCovered;
  for (unsigned int k1 = 0; k1 < aList.size(); k1++)
    aTracks[aList[k1].mIndex].mLabel = -1;
}

class CMergeType {
public:
  CMergeType(float aScore, int a, int b) : mScore(aScore),ma(a),mb(b) {}
  int ma,mb;
  float mScore;
};

bool operator<(const CMergeType& a, const CMergeType& b) {
  return a.mScore < b.mScore;
}

float segmentGraphMultiRegion(CMatrix<float>& aFeatures, CVector<float>& aWeights, int& aRegionNo, CVector<int>& aIndexToNode, CVector<int>& aNodeToIndex, std::vector<CTrack>& aTracks, CVector<int>& aResult) {
  float aNu = 8.0f;     
  float invNu = 1.0f/aNu;
  int N = aFeatures.xSize();
  int aDim = aFeatures.ySize();
  float aWeightNorm = 0.25f;
  CVector<float> invWeights(aDim);
  for (int d = 0; d < aDim; d++)
    invWeights(d) = 1.0f/(aWeightNorm*aWeights(d));
  CVector<int> aLabels(N);
  CVector<float>* aData = new CVector<float>[N];
  for (int n = 0; n < N; n++) {
    aData[n].setSize(aDim);
    for (int i = 0; i < aDim; i++)
      aData[n](i) = aFeatures(n,i)*aWeightNorm*aWeights(i);
  }
  // Some precomputations to quickly access grid information
  CVector<int> aLength(N);
  CVector<std::vector<float> > g(N);
  CVector<std::vector<int> > idx(N);
  CVector<float> w(N,0);
  int aNeighborCount = aTracks[aNodeToIndex(0)].mNearestNeighbors.size();
  for (int n = 0; n < N; n++) {
    g(n).reserve(2*aNeighborCount);
    idx(n).reserve(2*aNeighborCount);
  }
  // Compute edge weights; make sure that neighborhood relationship is symmetric
  for (int n = 0; n < N; n++) {
    int i = aNodeToIndex(n);
    aLength(n) = aTracks[i].mx.size();
    for (int r = 0; r < aNeighborCount; r++) {
      int j = aTracks[i].mNearestNeighbors(r);
      if (j < 0) continue;
      j = aIndexToNode(j);
      if (j < 0) continue;
      bool aFound = false;
      for (unsigned int a = 0; a < idx(n).size(); a++)
        if (idx(n)[a] == j) { aFound = true; break; }
      if (!aFound) {
        idx(n).push_back(j);
        float mag = sqrt(aTracks[i].mNearestNeighborsDX(r)*aTracks[i].mNearestNeighborsDX(r)+aTracks[i].mNearestNeighborsDY(r)*aTracks[i].mNearestNeighborsDY(r));
        float val = 1e-6f;
        for (int d = 0; d < aDim; d++) {
          float temp = (aData[n](d)-aData[j](d))*invWeights(d);
          val += temp*temp;
        }
        val = mag/(val);
        g(n).push_back(val);
        w(n) += val;
        for (unsigned int a = 0; a < idx(j).size(); a++)
          if (idx(j)[a] == n) { aFound = true; break; }
        if (!aFound) {
          idx(j).push_back(n);
          g(j).push_back(val);
          w(j) += val;
        }
      }
    }
  }
  for (int n = 0; n < N; n++)
    for (unsigned int r = 0; r < idx(n).size(); r++) {
      float max = NMath::max(w(n),w(idx(n)[r]));
      if (max > 0) g(n)[r] /= max;
    }
  //std::ofstream aOut((mResultDir+"Energies.txt").c_str());
  //aOut << "Eigenvalues:" << std::endl;
  //for (int d = 0; d < aDim; d++)
  //  aOut << aWeights(d) << " ";
  //aOut << std::endl;
  // Compute energy for one single region
  int bestRegionNo = 1;
  aResult.setSize(N); aResult = 0;
  float bestEnergy = 0.0f;
  CVector<float> aMean1(aDim,0);
  int sum = 0;
  for (int n = 0; n < N; n++) {
    aMean1 += ((float)aLength(n)*aData[n]);
    sum += aLength(n);
  }
  aMean1 *= 1.0f/sum;
  for (int n = 0; n < N; n++) {
    float dist = 0.0f;
    for (int i = 0; i < aDim; i++) {
      float temp = aData[n](i)-aMean1(i);
      dist += temp*temp;
    }
    bestEnergy += aLength(n)*dist;
  }
  //aOut << "1: " << bestEnergy << std::endl;
  int aNoImprovement = 0;
  // Now try more regions ------------------------------------------------------
  for (int K = 2; K <= aRegionNo; K++) {
    if (aNoImprovement > 5) break;
    CVector<float>* aSum = new CVector<float>[K];
    CVector<float>* aMean = new CVector<float>[K];
    float* aNum = new float[K];
    for (int k = 0; k < K; k++) {
      aSum[k].setSize(aDim);
      aMean[k].setSize(aDim);
    }
    CMatrix<float> aWork(N,K);
    // Repeat segmentation for several random initializations
    for (int t = 0; t < 1; t++) {
      // 1. K-Means iterations to quickly find a good initial segmentation -----
      kmeans(aData,aMean,aNum,aLabels,aLength,g,idx,K,aNu,aNodeToIndex,aTracks,mXSize,mYSize,mSequenceLength);
      #ifdef SHOWAFTERKMEANS
      for (int n = 0; n < N; n++)
        aTracks[aNodeToIndex(n)].mLabel = aLabels(n);
      for (int t = 0; t < mSequenceLength; t+=10)
        showSegments(aTracks,t,100+K,true);
      #endif
      // 2. Merging moves when there is not enough edge evidence along cluster boundary
      bool aPositiveMergeFound;
      do {
        float bestMergeEnergy = computeEnergy(aData,aMean,aLabels,aLength,g,idx,aNu);
        //aOut << "none: " << bestMergeEnergy << " ... ";
        CVector<int> bestLabels(aLabels);
        // Compute average gradient magnitude of eigenvectors between clusters
        CMatrix<float> aGradient(K,K,0.0f);
        CMatrix<float> aNeighborCount(K,K,0.0f);
        for (int n = 0; n < N; n++)
          for (unsigned int r = 0; r < idx(n).size(); r++) {
            int a = aLabels(n);
            int b = aLabels(idx(n)[r]);
            if (a != b) {
              aGradient(a,b) += aLength(n)*g(n)[r];
              aNeighborCount(a,b) += aLength(n);
              aGradient(b,a) += aLength(n)*g(n)[r];
              aNeighborCount(a,b) += aLength(n);
            }
          }
        std::priority_queue<CMergeType> aQueue;
        for (int a = 0; a < K; a++)
          for (int b = a+1; b < K; b++)
            if (aNeighborCount(a,b) > 0)
              aQueue.push(CMergeType(aGradient(a,b)/aNeighborCount(a,b),a,b));
        CVector<float>* tempSum = new CVector<float>[K];
        CVector<float>* tempMean = new CVector<float>[K];
        for (int k = 0; k < K; k++) {
          tempSum[k].setSize(aDim);
          tempMean[k].setSize(aDim);
        }
        float* tempNum = new float[K];
        int aMoveCount = 0;
        aPositiveMergeFound = false;
        while (!aQueue.empty() && aMoveCount < 10) {
          CMergeType aMove = aQueue.top(); aQueue.pop();
          CVector<int> tempLabels(aLabels);
          for (int n = 0; n < N; n++)
            if (tempLabels(n) == aMove.mb) tempLabels(n) = aMove.ma;
          kmeansRecomputeStatistics(aData,tempSum,tempNum,tempLabels,aLength,K);
          for (int k = 0; k < K; k++) {
            tempMean[k] = tempSum[k];
            if (tempNum[k] > 0) tempMean[k] *= 1.0f/tempNum[k];
          }
          float aEnergy = computeEnergy(aData,tempMean,tempLabels,aLength,g,idx,aNu);
          //aOut << aMove.ma << "," << aMove.mb << ": " << aEnergy << " ... ";
          if (aEnergy < bestMergeEnergy) {
            bestMergeEnergy = aEnergy;
            bestLabels = tempLabels;
            aPositiveMergeFound = true;
          }
          aMoveCount++;
        }
        delete[] tempSum;
        delete[] tempMean;
        delete[] tempNum;
        aLabels = bestLabels;
        if (aPositiveMergeFound) {
          kmeansRecomputeStatistics(aData,aSum,aNum,aLabels,aLength,K);
          for (int k = 0; k < K; k++) {
            aMean[k] = aSum[k];
            if (aNum[k] > 0) aMean[k] *= 1.0f/aNum[k];
          }
        }
        //aOut << std::endl;
      } while (aPositiveMergeFound);
      #ifdef SHOWAFTERMERGING
      for (int n = 0; n < N; n++)
        aTracks[aNodeToIndex(n)].mLabel = aLabels(n);
      for (int t = 0; t < mSequenceLength; t+=10)
        showSegments(aTracks,t,200+K,true);
      #endif
      // 3. Segmentation with smoothness prior ---------------------------------
      aWork = 0.0f;
      for (int n = 0; n < N; n++)
        aWork(n,aLabels(n)) = 1.0f;
      for (int t2 = 0; t2 < 10; t2++) {
        // 3a) Smoothing
        float aOmega = 1.85;
        float negOmega = 1.0f-aOmega;
        CMatrix<float> aSmoothed(aWork);
        for (int k = 0; k < K; k++) {
          float* pWork = aWork.data()+k*N;
          float* pSmoothed = aSmoothed.data()+k*N;
          for (int sor = 0; sor < 20; sor++)
            for (int n = 0; n < N; n++) {
              float sum = 0.0f;
              float num = 0.0f;
              for (unsigned int r = 0; r < g(n).size(); r++) {
                int j = idx(n)[r];
                float a = pSmoothed[j]-pSmoothed[n];
                float grad = a*a+1e-6f;
                float diff = g(n)[r]/sqrt(grad);
                sum += diff*pSmoothed[j];
                num += diff;
              }
              pSmoothed[n] = negOmega*pSmoothed[n]+aOmega*(sum+invNu*pWork[n])/(num+invNu);
            }
          for (int n = 0; n < N; n++)
            pSmoothed[n] -= pWork[n];
        }
        // 3b) Competition
        for (int k = 0; k < K; k++)
          for (int n = 0; n < N; n++) {
            float val = 0.0f;
            for (int i = 0; i < aDim; i++) {
              float temp = aData[n](i)-aMean[k](i);
              val -= temp*temp;
            }
            val = invNu*val+aSmoothed(n,k);
            float max = val-1000.0;
            for (int k2 = 0; k2 < K; k2++) {
              if (k == k2) continue;
              float val2 = 0.0f;
              for (int i = 0; i < aDim; i++) {
                float temp = aData[n](i)-aMean[k2](i);
                val2 -= temp*temp;
              }
              val2 = invNu*val2+aSmoothed(n,k2);
              if (val2 > max) max = val2;
            }
            aWork(n,k) += 0.5f*(val-max);
          }
        // 3c) Clipping
        for (int i = 0; i < aWork.size(); i++)
          if (aWork.data()[i] < 0.0f) aWork.data()[i] = 0.0f;
          else if (aWork.data()[i] > 1.0f) aWork.data()[i] = 1.0f;
        // 3d) Recompute means (with soft assignment)
        for (int k = 0; k < K; k++) {
          aSum[k] = 0.0f;
          aNum[k] = 0.0f;
          for (int n = 0; n < N; n++) {
            float w = aWork(n,k)*aLength(n);
            aSum[k] += w*aData[n];
            aNum[k] += w;
          }
          aMean[k] = aSum[k];
          if (aNum[k] > 0) aMean[k] *= 1.0f/aNum[k];
        }
        //if (t2 % 5 == 0) showWork(aWork,aNodeToIndex,aTracks,1);
      }
      // 4. Binarize -----------------------------------------------------------
      for (int n = 0; n < N; n++) {
        float max = aWork(n,0);
        aLabels(n) = 0;
        for (int k = 1; k < K; k++)
          if (aWork(n,k) > max) {
            max = aWork(n,k);
            aLabels(n) = k;
          }
      }
      // Recompute means
      kmeansRecomputeStatistics(aData,aSum,aNum,aLabels,aLength,K);
      for (int k = 0; k < K; k++) {
        aMean[k] = aSum[k];
        if (aNum[k] > 0) aMean[k] *= 1.0f/aNum[k];
      }
      float aEnergy = computeEnergy(aData,aMean,aLabels,aLength,g,idx,aNu);
      if (aEnergy < bestEnergy) {
        aResult = aLabels;
        bestEnergy = aEnergy;
        bestRegionNo = K;
        aNoImprovement = 0;
      }
      else aNoImprovement++;
      //aOut << K << ": " << aEnergy << " ";
    }
    //aOut << std::endl;
    delete[] aNum;
    delete[] aSum;
    delete[] aMean;
  }
  // Check for empty or tiny regions and rearrange labels if necessary
  for (int k = 0; k < bestRegionNo; k++) {
    int sum = 0;
    for (int n = 0; n < N; n++)
      if (aResult(n) == k) sum += aLength(n);
    if (sum < 1) {
      for (int n = 0; n < N; n++)
        if (aResult(n) == k) aResult(n) = -1;
      if (k < bestRegionNo-1) {
        for (int n = 0; n < N; n++)
          if (aResult(n) == bestRegionNo-1) aResult(n) = k;
        k--;
      }
      bestRegionNo--;
    }
  }
  delete[] aData;
  aRegionNo = bestRegionNo;
  return bestEnergy/N;
}

// splitTracks
void splitTracks(int aLabel, std::vector<CTrack>& aTracks, int& aRegionNo, int aLevel) {
  // Select subset that should be split
  CVector<bool> aValidTracks(aTracks.size());
  CVector<int> aTrackToNode(aTracks.size());
  int ai = 0;
  for (unsigned int i = 0; i < aTracks.size(); i++)
    if (aTracks[i].mLabel == aLabel) {
      aValidTracks(i) = true;
      aTrackToNode(i) = ai++;
    }
    else {
      aValidTracks(i) = false;
      aTrackToNode(i) = -1;
    }
  CVector<int> aNodeToTrack(ai);
  ai = 0;
  for (unsigned int i = 0; i < aTracks.size(); i++)
    if (aTracks[i].mLabel == aLabel) aNodeToTrack(ai++) = i;
  // Compute eigenvectors for this subset
  CVector<int> X;
  CMatrix<float> eigVecs;
  mStatus << "Before N-Cut." << std::endl;
  ncut(mGraph,aValidTracks,eigVecs,mEigVals,2);
  mStatus << "After N-Cut." << std::endl;
  // Normalize eigenvectors to have a dynamic range of 1
  for (int j = 0; j < eigVecs.ySize(); j++) {
    float min = 1e8;
    float max = -1e8;
    for (int i = 0; i < eigVecs.xSize(); i++) {
      if (eigVecs(i,j) > max) max = eigVecs(i,j);
      if (eigVecs(i,j) < min) min = eigVecs(i,j);
    }
    float aFactor = 1.0f/(max-min+0.0001*fabs(max));
    for (int i = 0; i < eigVecs.xSize(); i++)
      eigVecs(i,j) = (eigVecs(i,j)-min)*aFactor;
  }
  // Sort out eigenvectors that separate only tiny regions
  CVector<float> w(eigVecs.xSize());
  for (int n = 0; n < eigVecs.xSize(); n++)
    w(n) = sqrt(aTracks[aNodeToTrack(n)].mx.size());
  CVector<bool> aKeepVector(eigVecs.ySize(),true);
  int aNewVectorCount = eigVecs.ySize();
  CVector<bool> aKillTheseNodes(eigVecs.xSize(),false);
  for (int j = 0; j < eigVecs.ySize(); j++) {
    float sum[2];
    float num[2];
    CVector<char> aLabels(eigVecs.xSize());
    do {
      sum[0] = 0.0f; sum[1] = 0.0f; num[0] = 0.0f; num[1] = 0.0f;
      for (int n = 0; n < eigVecs.xSize(); n++) {
        aLabels(n) = (int)(2.0f*NMath::random());
        if (aLabels(n) >= 2) aLabels(n) = 1;
        sum[aLabels(n)] += w(n)*eigVecs(n,j);
        num[aLabels(n)] += w(n);
      }
    } while (num[0] <= 0 || num[1] <= 0);
    float mean[2];
    mean[0] = sum[0]/num[0];
    mean[1] = sum[1]/num[1];
    for (int i = 0; i < 10; i++) {
      int aChanged = 0;
      for (int n = 0; n < eigVecs.xSize(); n++) {
        char bestLabel;
        float diff0 = eigVecs(n,j)-mean[0];
        float diff1 = eigVecs(n,j)-mean[1];
        if (diff0*diff0 < diff1*diff1) bestLabel = 0;
        else bestLabel = 1;
        if (aLabels(n) != bestLabel) {
          char oldLabel = aLabels(n);
          sum[oldLabel] -= w(n)*eigVecs(n,j);
          num[oldLabel] -= w(n);
          mean[oldLabel] = sum[oldLabel];
          if (num[oldLabel] > 0.0f) mean[oldLabel] *= 1.0f/num[oldLabel];
          sum[bestLabel] += w(n)*eigVecs(n,j);
          num[bestLabel] += w(n);
          mean[bestLabel] = sum[bestLabel];
          mean[bestLabel] *= 1.0f/num[bestLabel];
          aLabels(n) = bestLabel;
          aChanged++;
        }
      }
      if (aChanged == 0) break;
    }
    if (num[0] < 100 || num[1] < 100) {
      aKeepVector(j) = false;
      aNewVectorCount--;
      char aLabelToBeRemoved = 0;
      if (num[1] < 100) aLabelToBeRemoved = 1;
      for (int n = 0; n < eigVecs.xSize(); n++)
        if (aLabels(n) == aLabelToBeRemoved) aKillTheseNodes(n) = true;
    }
  }
  CMatrix<float> oldVecs(eigVecs);
  CVector<float> oldVals(mEigVals);
  eigVecs.setSize(oldVecs.xSize(),aNewVectorCount);
  mEigVals.setSize(aNewVectorCount);
  int aj = 0;
  for (int j = 0; j < oldVecs.ySize(); j++)
    if (aKeepVector(j)) {
      mEigVals(aj) = oldVals(j);
      for (int n = 0; n < eigVecs.xSize(); n++)
        eigVecs(n,aj) = oldVecs(n,j);
      aj++;
    }
  for (unsigned int i = 0; i < aTracks.size(); i++) {
    aTracks[i].mProjection.setSize(eigVecs.ySize());
    if (aValidTracks(i) == false) aTracks[i].mProjection = -100;
    else {
      int ai = aTrackToNode(i);
      for (int k = 0; k < eigVecs.ySize(); k++)
        aTracks[i].mProjection(k) = eigVecs(ai,k);
    }
  }
  #ifdef SHOWEIGENVECTORS
  for (int t = mStartFrame; t < mStartFrame+mSequenceLength; t++)
    showEigenvectors(mTracks,t);
  #endif
  // Run segmentation (k-means with geodesic compactness)
  int newRegionNo = 2*mEigVals.size();
  mStatus << "Before segmentation." << std::endl;
  segmentGraphMultiRegion(eigVecs,mEigVals,newRegionNo,aTrackToNode,aNodeToTrack,aTracks,X);
  mStatus << "After segmentation." << std::endl;
  // Set new labels
  for (unsigned int i = 0; i < aTracks.size(); i++) {
    if (aValidTracks(i) == false) aTracks[i].mProjection = -100;
    else {
      int ai = aTrackToNode(i);
      if (X(ai) < 0 || aKillTheseNodes(ai)) aTracks[i].mLabel = -1;
      else if (X(ai) == 0) aTracks[i].mLabel = aLabel;
      else aTracks[i].mLabel = aRegionNo+X(ai)-1;
    }
  }
  // Add new regions to queue
  if (newRegionNo > 1) {
    mSplitQueue.push(aLabel);
    for (int k = 0; k < newRegionNo-1; k++)
      mSplitQueue.push(aRegionNo+k);
    aRegionNo += newRegionNo-1;
  }
}

void mergeClustersAffine(std::vector<CTrack>& aTracks) {
  // Determine which regions are touching each other
  CMatrix<bool> aNeighbors(mRegionNo,mRegionNo,false);
  for (unsigned int i = 0; i < aTracks.size(); i++) {
    int aLabelA = aTracks[i].mLabel;
    if (aLabelA < 0) continue;
    for (unsigned int j = i+1; j < aTracks.size(); j++) {
      int aLabelB = aTracks[j].mLabel;
      if (aLabelB < 0) continue;
      if (aNeighbors(aLabelA,aLabelB)) continue;
      for (int k = 0; k < aTracks[i].mNearestNeighbors.size(); k++) {
        int aIndex = aTracks[i].mNearestNeighbors(k);
        if (aIndex < 0) continue;
        if (aTracks[aIndex].mLabel < 0) continue;
        aNeighbors(aTracks[i].mLabel,aTracks[aIndex].mLabel) = true;
        aNeighbors(aTracks[aIndex].mLabel,aTracks[i].mLabel) = true;
      }
      for (int k = 0; k < aTracks[j].mNearestNeighbors.size(); k++) {
        int aIndex = aTracks[j].mNearestNeighbors(k);
        if (aIndex < 0) continue;
        if (aTracks[aIndex].mLabel < 0) continue;
        aNeighbors(aTracks[j].mLabel,aTracks[aIndex].mLabel) = true;
        aNeighbors(aTracks[aIndex].mLabel,aTracks[j].mLabel) = true;
      }
      CTrack& A = aTracks[i];
      CTrack& B = aTracks[j];
      int aStartTime = NMath::max(A.mSetupTime,B.mSetupTime);
      int aLength = NMath::min((int)A.mx.size()+A.mSetupTime-aStartTime,(int)B.mx.size()+B.mSetupTime-aStartTime);
      if (aLength <= 0) continue;
      float mindist = 1e10f;
      for (int t = aStartTime; t < aStartTime+aLength; t++) {
        float temp = A.mx[t-A.mSetupTime]-B.mx[t-B.mSetupTime];
        float dist = temp*temp;
        temp = A.my[t-A.mSetupTime]-B.my[t-B.mSetupTime];
        dist += temp*temp;
        if (dist < mindist) mindist = dist;
      }
      if (mindist < 4*mStep*mStep) {
        aNeighbors(aLabelA,aLabelB) = true;
        aNeighbors(aLabelB,aLabelA) = true;
      }
    }
  }
  mStatus << "Merging..." << std::endl;
  float min;
  int iter = 0;
  do {
    CMatrix<float> aError(mRegionNo,mSequenceLength,0.0f);
    CMatrix<int> aErrorCount(mRegionNo,mSequenceLength,0);
    CTensor<float> aLossA(mRegionNo,mRegionNo,mSequenceLength,0.0f);
    CTensor<float> aLossB(mRegionNo,mRegionNo,mSequenceLength,0.0f);
    char buffer[200];
    CTensor<float> aShow;
    CTensor<float> aFlow(mXSize,mYSize,2);
    for (int t = 0; t < mSequenceLength; t++) {
      // Estimate each cluster's size at this frame
      CVector<int> aRegionSizes(mRegionNo,0);
      for (unsigned int i = 0; i < aTracks.size(); i++) {
        if (aTracks[i].mLabel < 0) continue;
        int aTime = t-aTracks[i].mSetupTime;
        if (aTime < 0 || aTime >= aTracks[i].mx.size()) continue;
        aRegionSizes(aTracks[i].mLabel)++;
      }
      // Estimate affine model for each cluster at this frame
      for (int k = 0; k < mRegionNo; k++) {
        if (aRegionSizes(k) < 3) continue;
        // Estimate affine model
        CMatrix<float> A(6,2*aRegionSizes(k));
        CVector<float> b(2*aRegionSizes(k));
        float* pA = A.data();
        float* pb = b.data();
        for (unsigned int i = 0; i < aTracks.size(); i++) {
          if (aTracks[i].mLabel != k) continue;
          int aTime = t-aTracks[i].mSetupTime;
          if (aTime < 0 || aTime >= aTracks[i].mx.size()) continue;
          float x,y;
          if (aTime == 0) { x = aTracks[i].mox; y = aTracks[i].moy;}
          else { x = aTracks[i].mx[aTime-1]; y = aTracks[i].my[aTime-1];}
          *(pA++) = x;    *(pA++) = y;    *(pA++) = 0.0f;  *(pA++) = 0.0f;  *(pA++) = 1.0f; *(pA++) = 0.0f; *(pb++) = aTracks[i].mx[aTime];
          *(pA++) = 0.0f; *(pA++) = 0.0f; *(pA++) = x;     *(pA++) = y;     *(pA++) = 0.0f; *(pA++) = 1.0f; *(pb++) = aTracks[i].my[aTime];
        }
        CVector<float> X = NMath::leastSquares(A,b);
        #if 0
        if (k == 0 || k == 5 || k == 6) {
        for (int y = 0; y < mYSize; y++)
          for (int x = 0; x < mXSize; x++) {
            aFlow(x,y,0) = X(0)*x+X(1)*y+X(4)-x;
            aFlow(x,y,1) = X(2)*x+X(3)*y+X(5)-y;
          }
        COpticFlow::flowToImage(aFlow,aShow);
        sprintf(buffer,"Affine%03d_%d.ppm",t,k);
        aShow.downsample(aShow.xSize()/4,aShow.ySize()/4);
        aShow.writeToPPM((mResultDir+buffer).c_str());
        }
        #endif
        // Compute error of this model
        for (unsigned int i = 0; i < aTracks.size(); i++) {
          if (aTracks[i].mLabel != k) continue;
          int aTime = t-aTracks[i].mSetupTime;
          if (aTime < 0 || aTime >= aTracks[i].mx.size()) continue;
          float x,y,flowvar;
          if (aTime == 0) { x = aTracks[i].mox; y = aTracks[i].moy; flowvar = aTracks[i].mFlowVar[aTime];}
          else { x = aTracks[i].mx[aTime-1]; y = aTracks[i].my[aTime-1]; flowvar = aTracks[i].mFlowVar[aTime]-aTracks[i].mFlowVar[aTime-1];}
          float temp = X(0)*x+X(1)*y+X(4)-aTracks[i].mx[aTime];
          float dist = temp*temp;
          temp = X(2)*x+X(3)*y+X(5)-aTracks[i].my[aTime];
          dist += temp*temp;
          dist *= 1.0f/(flowvar*flowvar);
          aError(k,t) += dist;
          aErrorCount(k,t)++;
        }
      }
      // Estimate affine model for each cluster combination at this frame
      for (int k1 = 0; k1 < mRegionNo; k1++)
        for (int k2 = k1+1; k2 < mRegionNo; k2++) {
          if (aRegionSizes(k1) <= 3 || aRegionSizes(k2) <= 3) continue;
          if (aNeighbors(k1,k2) == false) continue;
          float wa = sqrt((aRegionSizes(k1)+aRegionSizes(k2))/aRegionSizes(k1));
          float wb = sqrt((aRegionSizes(k1)+aRegionSizes(k2))/aRegionSizes(k2));
          // Estimate affine model
          CMatrix<float> A(6,2*(aRegionSizes(k1)+aRegionSizes(k2)));
          CVector<float> b(2*(aRegionSizes(k1)+aRegionSizes(k2)));
          float* pA = A.data();
          float* pb = b.data();
          for (unsigned int i = 0; i < aTracks.size(); i++) {
            if (aTracks[i].mLabel != k1 && aTracks[i].mLabel != k2) continue;
            int aTime = t-aTracks[i].mSetupTime;
            if (aTime < 0 || aTime >= aTracks[i].mx.size()) continue;
            float x,y;
            if (aTime == 0) { x = aTracks[i].mox; y = aTracks[i].moy;}
            else { x = aTracks[i].mx[aTime-1]; y = aTracks[i].my[aTime-1];}
            if (aTracks[i].mLabel == k1) {
              *(pA++) = wa*x; *(pA++) = wa*y; *(pA++) = 0.0f; *(pA++) = 0.0f; *(pA++) = wa;   *(pA++) = 0.0f; *(pb++) = wa*aTracks[i].mx[aTime];
              *(pA++) = 0.0f; *(pA++) = 0.0f; *(pA++) = wa*x; *(pA++) = wa*y; *(pA++) = 0.0f; *(pA++) = wa;   *(pb++) = wa*aTracks[i].my[aTime];
            }
            else {
              *(pA++) = wb*x; *(pA++) = wb*y; *(pA++) = 0.0f; *(pA++) = 0.0f; *(pA++) = wb;   *(pA++) = 0.0f; *(pb++) = wb*aTracks[i].mx[aTime];
              *(pA++) = 0.0f; *(pA++) = 0.0f; *(pA++) = wb*x; *(pA++) = wb*y; *(pA++) = 0.0f; *(pA++) = wb;   *(pb++) = wb*aTracks[i].my[aTime];
            }
          }
          CVector<float> X = NMath::leastSquares(A,b);
          #if 0
          if (k1 == 0 && (k2 == 6 || k2 == 5)) {
          for (int y = 0; y < mYSize; y++)
            for (int x = 0; x < mXSize; x++) {
              aFlow(x,y,0) = X(0)*x+X(1)*y+X(4)-x;
              aFlow(x,y,1) = X(2)*x+X(3)*y+X(5)-y;
            }
          COpticFlow::flowToImage(aFlow,aShow);
          sprintf(buffer,"Affine%03d_%d_%d.ppm",t,k1,k2);
          aShow.downsample(aShow.xSize()/4,aShow.ySize()/4);
          aShow.writeToPPM((mResultDir+buffer).c_str());
          }
          #endif
          // Compute error of this model
          int aCount = 0;
          for (unsigned int i = 0; i < aTracks.size(); i++) {
            if (aTracks[i].mLabel != k1 && aTracks[i].mLabel != k2) continue;
            int aTime = t-aTracks[i].mSetupTime;
            if (aTime < 0 || aTime >= aTracks[i].mx.size()) continue;
            float x,y,flowvar;
            if (aTime == 0) { x = aTracks[i].mox; y = aTracks[i].moy; flowvar = aTracks[i].mFlowVar[aTime];}
            else { x = aTracks[i].mx[aTime-1]; y = aTracks[i].my[aTime-1]; flowvar = aTracks[i].mFlowVar[aTime]-aTracks[i].mFlowVar[aTime-1];}
            float temp = X(0)*x+X(1)*y+X(4)-aTracks[i].mx[aTime];
            float dist = temp*temp;
            temp = X(2)*x+X(3)*y+X(5)-aTracks[i].my[aTime];
            dist += temp*temp;
            dist *= 1.0f/(flowvar*flowvar);
            if (aTracks[i].mLabel == k1) aLossA(k1,k2,t) += dist;
            else aLossB(k1,k2,t) += dist;
            aCount++;
          }
        }
    }
    min = 1e8;
    int mina,minb;
    //std::ofstream aOut((mResultDir+"out.txt").c_str());
    for (int a = 0; a < mRegionNo; a++)
      for (int b = a+1; b < mRegionNo; b++) {
        if (aNeighbors(a,b) == false) { /*aOut << a << "," << b << " Not connected." << std::endl;*/ continue;}
        float max = 0.0f;
        float maxloss = 0.0f;
        int maxT = -1;
        for (int t = 0; t < mSequenceLength; t++)
          if (aErrorCount(a,t) > 3 && aErrorCount(b,t) > 3) {
            float lossA = (aLossA(a,b,t)-aError(a,t))/aErrorCount(a,t);
            float lossB = (aLossB(a,b,t)-aError(b,t))/aErrorCount(b,t);
            float loss = NMath::max(lossA,lossB);
            int count = NMath::min(aErrorCount(a,t),aErrorCount(b,t));
            //aOut << a << "," << b << " -- " << t << ": " << count*loss << " " << loss << " " << lossA << " " << lossB << " " << aErrorCount(a,t) << " " << aErrorCount(b,t) << std::endl;
            if (loss*count > max) {
              max = loss*count;
              maxloss = loss;
              maxT = t;
            }
          }
        if (maxT < 0) { /*aOut << a << "," << b << " No common frames." << std::endl;*/ continue; }
        if (maxloss < min) {
          min = maxloss;
          mina = a;
          minb = b;
        }
      }
    if (min > 2.0f) break;
    for (unsigned int i = 0; i < aTracks.size(); i++)
      if (aTracks[i].mLabel == minb) aTracks[i].mLabel = mina;
    for (int c = 0; c < mRegionNo; c++)
      if (aNeighbors(c,minb)) {
        aNeighbors(mina,c) = true;
        aNeighbors(c,mina) = true;
      }
    mStatus << "Merging " << mina << " and " << minb << ": " << min << std::endl;
    iter++;
  } while (true);
}

// clustering
void clustering(std::vector<CTrack>& aTracks) {
  CVector<bool> aValidTracks(aTracks.size(),true);
  // Discard empty tracks
  for (unsigned int i = 0; i < aTracks.size(); i++) {
    aTracks[i].mLabel = -1;
    if (aTracks[i].mx.size() < 1) aValidTracks(i) = false;
  }
  // Build the graph
  buildGraph(aTracks,aValidTracks,mGraph);
  mRegionNo = 1;
  splitTracks(0,aTracks,mRegionNo,0);
  mergeClustersAffine(aTracks);
}

int main(int argc, char** args) {
  // Set memory limit to prevent crashing nodes on the cluster
  rlimit aLimit;
  aLimit.rlim_cur = 6000000000;
  aLimit.rlim_max = 7000000000;
  setrlimit(RLIMIT_DATA,&aLimit);
  // Determine input directory
  std::string s = args[1];
  s.erase(s.find_last_of("/")+1,s.length());
  mInputDir = s;
  s = args[1];
  s.erase(0,s.find_last_of('.'));
  // Read image sequence in bmf format
  if (s == ".bmf" || s == ".BMF") {
    int aImageCount,aViewCount;
    std::ifstream aStream(args[1]);
    aStream >> aImageCount;
    aStream >> aViewCount;
    mInput.setSize(aImageCount);
    for (int i = 0; i < aImageCount; i++) {
      std::string s;
      aStream >> s;
      mInput(i) = mInputDir+s;
    }
  }
  else {
    std::cout << "Must pass a bmf file as input" << std::endl;
    return -1;
  }
  // Determine image/sequence name
  s = args[1];
  s.erase(0,s.find_last_of("/")+1);
  s.erase(s.find_first_of('.'));
  mFilename = s;
  // Make directory for results
  mResultDir = mInputDir + "BroxMalikResults";
  std::string s2 = "mkdir " + mResultDir;
  int dummy = system(s2.c_str());
  mResultDir += "/";
  buildColorCode();
  // Set beginning and end of tracking
  mStartFrame = atoi(args[2]);
  mSequenceLength = atoi(args[3]);
  mStep = atoi(args[4]);
  char buffer[100];
  sprintf(buffer,"Status%d.txt",mSequenceLength);
  mStatus.open((mResultDir+buffer).c_str());
  clock_t start = clock();
  // Load first image
  CTensor<float>* aImage1 = new CTensor<float>;
  CTensor<float>* aImage2;
  aImage1->readFromPPM(mInput(mStartFrame).c_str());
  mXSize = aImage1->xSize();
  mYSize = aImage1->ySize();
  // Read images
  if (readTrackFromFile(mTracks) == false) {
    mStatus << "Tracks are being computed..." << std::endl;
    // Smooth first image
    NFilter::recursiveSmoothX(*aImage1,0.8f);
    NFilter::recursiveSmoothY(*aImage1,0.8f);
    for (int i = 0; i < aImage1->size(); i++) {
      if (aImage1->data()[i] > 255) aImage1->data()[i] = 255;
      if (aImage1->data()[i] < 0) aImage1->data()[i] = 0;
    }
    int aSize = mXSize*mYSize;
    // Compute HOG and HOG energy
    CTensor<float>* aHOG1 = new CTensor<float>;
    CTensor<float>* aHOG2;
    CMatching::computeHOG(*aImage1,*aHOG1);
    CMatrix<float>* aEnergy1 = new CMatrix<float>(mXSize,mYSize,0);
    CMatrix<float>* aEnergy2;
    int i2 = 0;
    for (int z = 0; z < aHOG1->zSize(); z++)
      for (int i = 0; i < aSize; i++,i2++)
        aEnergy1->data()[i] += aHOG1->data()[i2];
    // Initialize tracks
    mTracks.clear();
    CMatrix<float> aCorners;
    CMatrix<float> aCovered(mXSize,mYSize);
    // Tracking
    for (int t = 0; t < mSequenceLength-1; t++) {
      // Load next image
      aImage2 = new CTensor<float>;
      aImage2->readFromPPM(mInput(mStartFrame+t+1).c_str());
      NFilter::recursiveSmoothX(*aImage2,0.8f);
      NFilter::recursiveSmoothY(*aImage2,0.8f);
      for (int i = 0; i < aImage2->size(); i++) {
        if (aImage2->data()[i] > 255) aImage2->data()[i] = 255;
        if (aImage2->data()[i] < 0) aImage2->data()[i] = 0;
      }
      // Compute HOG and HOG energy
      aHOG2 = new CTensor<float>;
      CMatching::computeHOG(*aImage2,*aHOG2);
      aEnergy2 = new CMatrix<float>(mXSize,mYSize,0);
      float aSum = 0.0f;
      int i2 = 0;
      for (int z = 0; z < aHOG1->zSize(); z++)
        for (int i = 0; i < aSize; i++,i2++) {
          aEnergy2->data()[i] += aHOG2->data()[i2];
          aSum += aHOG2->data()[i2];
        }
      // Mark areas sufficiently covered by tracks
      aCovered = 1e20;
      if (t > 0) {
        for (unsigned int i = 0; i < mTracks.size(); i++)
          if (mTracks[i].mStopped == false) aCovered((int)mTracks[i].mx.back(),(int)mTracks[i].my.back()) = 0.0f;
        euclideanDistanceTransform(aCovered);
      }
      // Set up new tracking points in uncovered areas
      CMatching::computeCorners(*aImage1,aCorners,1.0f);
      float aCornerAvg = aCorners.avg();
      for (int ay = 4; ay < mYSize-4; ay+=mStep)
        for (int ax = 4; ax < mXSize-4; ax+=mStep) {
          if (aCovered(ax,ay) < mStep*mStep) continue;
          float distToImageBnd = exp(-0.1*NMath::min(NMath::min(NMath::min(ax,ay),mXSize-ax),mYSize-ay));
          if (aCorners(ax,ay) < aCornerAvg*(0.1+distToImageBnd)) continue;
          if (aCorners(ax,ay) < (1.0f+distToImageBnd)) continue;
          mTracks.push_back(CTrack());
          CTrack& newTrack = mTracks.back();
          newTrack.mox = ax;
          newTrack.moy = ay;
          newTrack.mLabel = -1;
          newTrack.mSetupTime = t;
          int tempx = ax-4;
          int tempy = ay-4;
          if ((tempx & 1) == 1 || (tempy & 1) == 1) newTrack.mScale = 1;
          else if ((tempx & 2) == 2 || (tempy & 2) == 2) newTrack.mScale = 2;
          else if ((tempx & 4) == 4 || (tempy & 4) == 4) newTrack.mScale = 4;
          else if ((tempx & 8) == 8 || (tempy & 8) == 8) newTrack.mScale = 8;
          else if ((tempx & 16) == 16 || (tempy & 16) == 16) newTrack.mScale = 16;
          else newTrack.mScale = 32;
        }
      // Show
      CTensor<float> aShow;
      if (t == 0) {
        aShow = *aImage1;
        for (unsigned int i = 0; i < mTracks.size(); i++) {
          int iax = mTracks[i].mox;
          int iay = mTracks[i].moy;
          aShow.drawRect(iax-1,iay-1,iax+1,iay+1,mColorCode(0,0),mColorCode(1,0),mColorCode(2,0));
        }
        sprintf(buffer,"Tracking%03d.ppm",mStartFrame);
        aShow.writeToPPM((mResultDir+buffer).c_str());
      }
      // Compute bidirectional HOG flow or read from file when available
      CTensor<float> aForward,aBackward;
      sprintf(buffer,"ForwardFlow%03d.flo",t+mStartFrame);
      std::string aName1 = mResultDir+buffer;
      sprintf(buffer,"BackwardFlow%03d.flo",t+mStartFrame);
      std::string aName2 = mResultDir+buffer;
      if (COpticFlow::readMiddlebury(aName1.c_str(),aForward,false) == false || COpticFlow::readMiddlebury(aName2.c_str(),aBackward,false) == false) {
        mStatus << "Compute optical flow. Frame" << t << std::endl;
        CMatching::hogFlow(*aImage1,*aImage2,*aHOG1,*aHOG2,*aEnergy1,*aEnergy2,aForward,aBackward);
        COpticFlow::writeMiddlebury(aForward,aName1.c_str());
        COpticFlow::writeMiddlebury(aBackward,aName2.c_str());
        //showFlow(aForward,aBackward,t);
      }
      CMatrix<float> aFlowVar;
      computeLocalFlowVariation(aForward,aFlowVar);
      // Check consistency of forward flow via backward flow
      CMatrix<float> aUnreliable(mXSize,mYSize,0);
      CTensor<float> dx(mXSize,mYSize,2);
      CTensor<float> dy(mXSize,mYSize,2);
      CDerivative<float> aDev(3);
      NFilter::filter(aForward,dx,aDev,1,1);
      NFilter::filter(aForward,dy,1,aDev,1);
      CMatrix<float> aMotionEdge(mXSize,mYSize,0);
      for (int i = 0; i < aSize; i++) {
        aMotionEdge.data()[i] += dx.data()[i]*dx.data()[i];
        aMotionEdge.data()[i] += dx.data()[aSize+i]*dx.data()[aSize+i];
        aMotionEdge.data()[i] += dy.data()[i]*dy.data()[i];
        aMotionEdge.data()[i] += dy.data()[aSize+i]*dy.data()[aSize+i];
      }
      for (int ay = 0; ay < aForward.ySize(); ay++)
        for (int ax = 0; ax < aForward.xSize(); ax++) {
          float bx = ax+aForward(ax,ay,0);
          float by = ay+aForward(ax,ay,1);
          int x1 = (int)bx;
          int y1 = (int)by;
          int x2 = x1+1;
          int y2 = y1+1;
          if (x1 < 0 || x2 >= mXSize || y1 < 0 || y2 >= mYSize) { aUnreliable(ax,ay) = 1.0f; continue;}
          float alphaX = bx-x1; float alphaY = by-y1;
          float a = (1.0-alphaX)*aBackward(x1,y1,0)+alphaX*aBackward(x2,y1,0);
          float b = (1.0-alphaX)*aBackward(x1,y2,0)+alphaX*aBackward(x2,y2,0);
          float u = (1.0-alphaY)*a+alphaY*b;
          a = (1.0-alphaX)*aBackward(x1,y1,1)+alphaX*aBackward(x2,y1,1);
          b = (1.0-alphaX)*aBackward(x1,y2,1)+alphaX*aBackward(x2,y2,1);
          float v = (1.0-alphaY)*a+alphaY*b;
          float cx = bx+u;
          float cy = by+v;
          float u2 = aForward(ax,ay,0);
          float v2 = aForward(ax,ay,1);
          if (((cx-ax)*(cx-ax)+(cy-ay)*(cy-ay)) >= 0.01*(u2*u2+v2*v2+u*u+v*v)+0.25f) { aUnreliable(ax,ay) = 1.0f; continue;}
          if (aMotionEdge(ax,ay) > 0.01*(u2*u2+v2*v2)+0.005f) { aUnreliable(ax,ay) = 1.0f; continue;}
          float distToImageBnd = exp(-0.1*NMath::min(NMath::min(NMath::min(ax,ay),mXSize-ax),mYSize-ay));
          if (aCorners(ax,ay) < 0.25f*aCornerAvg*(0.1f+distToImageBnd)) { aUnreliable(ax,ay) = 1.0f; continue;}
          if (aCorners(ax,ay) < 0.25f*(1.0f+distToImageBnd)) { aUnreliable(ax,ay) = 1.0f; continue;}
         }
      aShow = *aImage2;
      for (unsigned int i = 0; i < mTracks.size(); i++) {
        if (mTracks[i].mStopped) continue;
        float ax,ay,oldvar;
        if (mTracks[i].mSetupTime == t) {
          ax = mTracks[i].mox; ay = mTracks[i].moy; oldvar = 0.0f;
        }
        else {
          ax = mTracks[i].mx.back(); ay = mTracks[i].my.back(); oldvar = mTracks[i].mFlowVar.back();
        }
        int iax = lroundf(ax);
        int iay = lroundf(ay);
        if (aUnreliable(iax,iay) > 0) mTracks[i].mStopped = true;
        else {
          float bx = ax+aForward(iax,iay,0);
          float by = ay+aForward(iax,iay,1);
          int ibx = lroundf(bx);
          int iby = lroundf(by);
          if (ibx < 0 || iby < 0 || ibx >= mXSize || iby >= mYSize) mTracks[i].mStopped = true;
          else {
            mTracks[i].mx.push_back(bx);
            mTracks[i].my.push_back(by);
            mTracks[i].mFlowVar.push_back(oldvar+aFlowVar(iax,iay));
            int t = mTracks[i].mx.size();
            aShow.drawRect(ibx-1,iby-1,ibx+1,iby+1,mColorCode(0,4*t),mColorCode(1,4*t),mColorCode(2,4*t));
          }
        }
      }
      sprintf(buffer,"Tracking%03d.ppm",t+mStartFrame+1);
      aShow.writeToPPM((mResultDir+buffer).c_str());
      // Prepare for next image
      delete aImage1; aImage1 = aImage2;
      delete aHOG1; aHOG1 = aHOG2;
      delete aEnergy1; aEnergy1 = aEnergy2;
    }
    delete aImage1; delete aHOG1; delete aEnergy1;
    //writeTrackToFile(mTracks);
  }
  // Compute motion clusters
  clustering(mTracks);
  // Write results to files
  writeTracks();
  computeError(mTracks);
  for (int t = mStartFrame; t < mStartFrame+mSequenceLength; t++) {
    showSegments(mTracks,t);
  }
  clock_t finish = clock();
  mStatus << "Computation time: " << (double(finish)-double(start))/CLOCKS_PER_SEC << " seconds" << std::endl;
  mStatus.close();
  return 0;
}

